/* -*-C-*-
*******************************************************************************
*
* File:         avs_dsink.c
* RCS:          $Id: avs_dsink.c,v 1.1.1.1 2003/06/11 18:48:52 wyan Exp $
* Description:  
* Author:       Fabian E. Bustamante
*               Systems Research Group
*               College of Computing
*               Georgia Institute of Technology
* Created:      Mon Jun 18, 2001 at 11:23:58
* Modified:     Fri Oct 12, 2001 at 10:02:10 fabianb@cc.gatech.edu
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
* (C) Copyright 2001, Georgia Institute of Technology, all rights reserved.
*
*******************************************************************************
*/

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/types.h>  
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <string.h>

#include <io.h>
#include <atl.h>
#include <cm.h>
#include <echo.h>

#include <vdev.h>

#include <getopt.h>
#include <hrtime.h>
#include <tv.h>
#include <avs_util.h>
#include <avs_raw.h>
#include <styx.h>
#include <strlt.h>
#include <sensNet.h>
#include <stats.h>

#define SLOC_REMOTE 1
#define SLOC_LOCAL  2

static char* local_progname;
static boolean verbose = FALSE;		/* verbose mode */
static CManager cm;
static EChannel chan, derivedChan; 
static int NumRec;
static int NumRecReceived = 0;
static boolean MeasureMe = FALSE;
static int Freq = 2;
static char *StreamletName;
static int StreamletLoc = 0;
static struct timeval startTime, stopTime;
static struct timeval startCpu, stopCpu;
static struct timeval startUCpu, stopUCpu;
static struct timeval startSCpu, stopSCpu;
static char* default_strlt_path = "/home/fabianb/avs/strlt";

/* Command line arguments */
static struct option longopts[] =
{
  {"verbose", no_argument, NULL, 'v'},
  {"help", no_argument, NULL, 'h'},
  {"numframes", required_argument, NULL, 'n'},
  {"measure", no_argument, NULL, 'm'},
  {"sample", required_argument, NULL, 's'},
  {"transform", required_argument, NULL, 't'},
  {"placement", required_argument, NULL, 'p'},
  {"channel", required_argument, NULL, 'c'},
  {NULL, no_argument, NULL, 0}
};


/*----------------------------------------------------------------------------*/

/*
 * usage: Prints out 'usage' string
 *
 * Parameters: target 'stream' and 'status' to quit with
 *
 * Returns: -
 *
 */

static void 
usage (stream, status)
     FILE* stream; 
     int status;
{
  fprintf(stream, "Usage: %s [-vhm] -n nframes -s freq [-t name [-p place]] "
	  "-c chan-id\n"
	  "-v --verbose           \tverbose mode\n"
	  "-h --help              \tgive this help\n"
	  "-m --measure           \tset bandwidth measuring at client\n"
	  "-s --sample            \tsampling frequency for measurements\n"
	  "-n --numframes  num    \tnumber of expected frames (0 == infinite)\n"
	  "-t --transform  name   \ttransform stream with streamlet <name>\n"
	  "-p --placement  l|r    \tloc of streamlet (l: local, r: remote)\n"
	  "-c --channel chan-id   \tchannel id\n"
	  , local_progname);
  exit(status);
} /* end usage */


/*----------------------------------------------------------------------------*/

void
dumpStats()
{
  struct timeval usedCpu, usedUCpu, usedSCpu, usedTime;
  double *bwHist;

  tv_minus2(&usedTime, &stopTime, &startTime);
  tv_minus2(&usedCpu, &stopCpu, &startCpu);
  tv_minus2(&usedUCpu, &stopUCpu, &startUCpu);
  tv_minus2(&usedSCpu, &stopSCpu, &startSCpu);
  fprintf(stdout, "\n----------------------------------------\n");
  fprintf(stdout, "Sink performance data\n");
  fprintf(stdout, "---------------------\n");
  fprintf(stdout, "Total time:     %ld ms\n", tv_toMsecs(&usedTime)); 
  fprintf(stdout, "Time per event: %.3g ms\n", 
	  (float) tv_toMsecs(&usedTime)/NumRecReceived);
  fprintf(stdout, "CPU usage:      %ld ms (%.3g) \n", 
	  tv_toMsecs(&usedCpu), 
	  (float) tv_toMsecs(&usedCpu)/tv_toMsecs(&usedTime));
  fprintf(stdout, "User CPU usage:   %ld ms (%.3g)\n", 
	  tv_toMsecs(&usedUCpu), 
	  (float) tv_toMsecs(&usedUCpu)/tv_toMsecs(&usedTime));
  fprintf(stdout, "System CPU usage: %ld ms (%.3g)\n", 
	  tv_toMsecs(&usedSCpu), 
	  (float) tv_toMsecs(&usedSCpu)/tv_toMsecs(&usedTime));
  if (!strcmp(StreamletName, "halfRate") || (NumRec > 0))
      fprintf(stdout, "Percentage of events received (%d/%d):  %.3g\n", 
	      NumRecReceived, NumRec,(float) NumRecReceived/NumRec);

  if (MeasureMe) {
    bwHist = (double*) malloc (((int) (NumRecReceived/Freq)) * sizeof(double));
    if (bwHist == NULL) Styx_errQuit("Not enough resources");
    sensNet_GetHistory(bwHist, (int) (NumRecReceived/Freq), SENSNET_READ);
    fprintf(stdout, "Bandwidth mean:   \t%.3g Mbps (sdev %.3g)\n",
	    Stats_mean (bwHist, (int) (NumRecReceived/Freq)),
	    Stats_sdev (bwHist, (int) (NumRecReceived/Freq)));
    sensNet_Finish();
    free(bwHist);
  }
  fflush(stdout);
}

/*----------------------------------------------------------------------------*/

/*
 * getUsage: Get total CPU usage and wall-clock time point
 *
 * Parameters: Pointer to timeval structs for CPU and Time point
 * 
 * Return: Upon successful completion, 0 is returned.
 *
 */

int
getUsage(pointCpu, pointUCpu, pointSCpu, pointTime)
     struct timeval *pointCpu;
     struct timeval* pointUCpu;
     struct timeval* pointSCpu; 
     struct timeval *pointTime;
{
  struct rusage rusageBuffer;
  
  getrusage(RUSAGE_SELF, &rusageBuffer);
  tv_add2(pointCpu, &(rusageBuffer.ru_utime), &(rusageBuffer.ru_stime));
  tv_assign(pointUCpu, &(rusageBuffer.ru_utime));
  tv_assign(pointSCpu, &(rusageBuffer.ru_stime));
  gettimeofday(pointTime, NULL);
  
  return 0;
} /* end getUsage */


/*----------------------------------------------------------------------------*/

static void
interruptHandler(signo) 
     int signo;
{
  /* If the clock has not been stop after reaching the 
     target number of frames ... */
  if (NumRec == 0)
    getUsage(&stopCpu, &stopUCpu, &stopSCpu, &stopTime);

  dumpStats();  
  if (StreamletLoc == SLOC_REMOTE) EChannel_destroy(derivedChan);
  else EChannel_destroy(chan);
  CManager_close(cm);
  exit(0);
} /* end interruptHandler */


/*----------------------------------------------------------------------------*/

/*
 * eventHandler:
 *
 * Parameters: 
 * 
 * Return: 
 *
 */

void 
eventHandler (event, client_data, attrs)
     void* event; 
     void* client_data; 
     attr_list attrs;
{
  Raw_data_ptr rawrecP;
  Raw_data1C_ptr rawrec1CP; Raw_data1G_ptr rawrec1GP;   
  Raw_data1DC_ptr rawrec1DCP; Raw_data1DG_ptr rawrec1DGP;
  Raw_data2C_ptr rawrec2CP; Raw_data2G_ptr rawrec2GP;
  Raw_data2DC_ptr rawrec2DCP; Raw_data2DG_ptr rawrec2DGP;
  Raw_data3C_ptr rawrec3CP; Raw_data3G_ptr rawrec3GP;
  Raw_data3DC_ptr rawrec3DCP; Raw_data3DG_ptr rawrec3DGP;
  Raw_data4C_ptr rawrec4CP; Raw_data4G_ptr rawrec4GP;
  Raw_data4DC_ptr rawrec4DCP; Raw_data4DG_ptr rawrec4DGP;
  struct tm *now;
  time_t nowt;
  char snapshotFile[FILENAMELEN];
  char header[PPMHEADERLEN];
  int sfd;
  char *format;
  char ppm1 = '-', ppm2 = '-';
  int width = 0, height = 0, size = 0, tag = 0;
  unsigned char *tmpBuff = NULL;

  ++NumRecReceived;
  time(&nowt);
  now = localtime(&nowt);

  /* Untyped stream */
  if (!client_data) {
     rawrecP = (Raw_data_ptr) event;
     if (verbose) {
       fprintf(stdout, "Got event untyped %d\n", rawrecP->tag);
       fprintf(stdout, "             ppm1 %c\n", rawrecP->ppm1);
       fprintf(stdout, "             ppm2 %c\n", rawrecP->ppm2);
       fprintf(stdout, "             size %d\n", rawrecP->size);
     }
     
     sprintf(snapshotFile, "snapshot%d-%d-%d.ppm", 
	     NumRecReceived, now->tm_hour, now->tm_min);
     
     ppm1 = rawrecP->ppm1; ppm2 = rawrecP->ppm2;
     width = rawrecP->width; height = rawrecP->height;
     tmpBuff = rawrecP->buff; 
     size = rawrecP->size;
     tag = rawrecP->tag;

  } else {			/* Typed stream - which then? */
    format = (char*) client_data;
    if (!strcmp(format, "Raw_data1C\0")) {
      rawrec1CP = (Raw_data1C_ptr) event;
      if (verbose) {
	fprintf(stdout, "Got event type 1C %d\n", rawrec1CP->tag);
	fprintf(stdout, "             ppm1 %c\n", rawrec1CP->ppm1);
	fprintf(stdout, "             ppm2 %c\n", rawrec1CP->ppm2);
	fprintf(stdout, "             size %d\n", rawrec1CP->size);
      }
      
      sprintf(snapshotFile, "snapshot%d-%d-%d.ppm", 
	      NumRecReceived, now->tm_hour, now->tm_min);
      
      ppm1 = rawrec1CP->ppm1; ppm2 = rawrec1CP->ppm2;
      width = rawrec1CP->width; height = rawrec1CP->height;
      tmpBuff = rawrec1CP->buff; 
      size = rawrec1CP->size;
      tag = rawrec1CP->tag;
      
    } else if (!strcmp(format, "Raw_data2C\0")) {
      rawrec2CP = (Raw_data2C_ptr) event;
      if (verbose) {
	fprintf(stdout, "Got event type 1C %d\n", rawrec2CP->tag);
	fprintf(stdout, "             ppm1 %c\n", rawrec2CP->ppm1);
	fprintf(stdout, "             ppm2 %c\n", rawrec2CP->ppm2);
	fprintf(stdout, "             size %d\n", rawrec2CP->size);
      }
      sprintf(snapshotFile, "snapshot%d-%d-%d.ppm", 
	      NumRecReceived, now->tm_hour, now->tm_min);
      
      ppm1 = rawrec2CP->ppm1; ppm2 = rawrec2CP->ppm2;
      width = rawrec2CP->width; height = rawrec2CP->height;
      tmpBuff = rawrec2CP->buff; 
      size = rawrec2CP->size;
      tag = rawrec2CP->tag;
      
    } else if (!strcmp(format, "Raw_data3C\0")) {
      rawrec3CP = (Raw_data3C_ptr) event;
      if (verbose) fprintf(stdout, "Got event %d\n", rawrec3CP->tag);
      
      sprintf(snapshotFile, "snapshot%d-%d-%d.ppm", 
	      NumRecReceived, now->tm_hour, now->tm_min);
      
      ppm1 = rawrec3CP->ppm1; ppm2 = rawrec3CP->ppm2;
      width = rawrec3CP->width; height = rawrec3CP->height;
      tmpBuff = rawrec3CP->buff; 
      size = rawrec3CP->size;
      tag = rawrec3CP->tag;
      
    } else if (!strcmp(format, "Raw_data4C\0")) {
      rawrec4CP = (Raw_data4C_ptr) event;
      if (verbose) fprintf(stdout, "Got event %d\n", rawrec4CP->tag);
      
      sprintf(snapshotFile, "snapshot%d-%d-%d.ppm", 
	      NumRecReceived, now->tm_hour, now->tm_min);
      
      ppm1 = rawrec4CP->ppm1; ppm2 = rawrec4CP->ppm2;
      width = rawrec4CP->width; height = rawrec4CP->height;
      tmpBuff = rawrec4CP->buff; 
      size = rawrec4CP->size;
      tag = rawrec4CP->tag;
    } else if (!strcmp(format, "Raw_data1G\0")) {
      rawrec1GP = (Raw_data1G_ptr) event;
      if (verbose) fprintf(stdout, "Got event %d\n", rawrec1GP->tag);
      
      sprintf(snapshotFile, "snapshot%d-%d-%d.pgm", 
	      NumRecReceived, now->tm_hour, now->tm_min);
      
      ppm1 = rawrec1GP->ppm1; ppm2 = rawrec1GP->ppm2;
      width = rawrec1GP->width; height = rawrec1GP->height;
      tmpBuff = rawrec1GP->buff; 
      size = rawrec1GP->size;
      tag = rawrec1GP->tag;
    } else if (!strcmp(format, "Raw_data2G\0")) {
      rawrec2GP = (Raw_data2G_ptr) event;
      if (verbose) fprintf(stdout, "Got event %d\n", rawrec2GP->tag);
      
      sprintf(snapshotFile, "snapshot%d-%d-%d.pgm", 
	      NumRecReceived, now->tm_hour, now->tm_min);
      
      ppm1 = rawrec2GP->ppm1; ppm2 = rawrec2GP->ppm2;
      width = rawrec2GP->width; height = rawrec2GP->height;
      tmpBuff = rawrec2GP->buff; 
      size = rawrec2GP->size;
      tag = rawrec2GP->tag;
    } else if (!strcmp(format, "Raw_data3G\0")) {
      rawrec3GP = (Raw_data3G_ptr) event;
      if (verbose) fprintf(stdout, "Got event %d\n", rawrec3GP->tag);
      
      sprintf(snapshotFile, "snapshot%d-%d-%d.pgm", 
	      NumRecReceived, now->tm_hour, now->tm_min);
      
      ppm1 = rawrec3GP->ppm1; ppm2 = rawrec3GP->ppm2;
      width = rawrec3GP->width; height = rawrec3GP->height;
      tmpBuff = rawrec3GP->buff; 
      size = rawrec3GP->size;
      tag = rawrec3GP->tag;
    } else if (!strcmp(format, "Raw_data4G\0")) {
      rawrec4GP = (Raw_data4G_ptr) event;
      if (verbose) fprintf(stdout, "Got event %d\n", rawrec4GP->tag);
      
      sprintf(snapshotFile, "snapshot%d-%d-%d.pgm", 
	      NumRecReceived, now->tm_hour, now->tm_min);
      
      ppm1 = rawrec4GP->ppm1; ppm2 = rawrec4GP->ppm2;
      width = rawrec4GP->width; height = rawrec4GP->height;
      tmpBuff = rawrec4GP->buff; 
      size = rawrec4GP->size;
      tag = rawrec4GP->tag;
    } else if (!strcmp(format, "Raw_data1DC\0")) {
      rawrec1DCP = (Raw_data1DC_ptr) event;
      if (verbose) fprintf(stdout, "Got event %d\n", rawrec1DCP->tag);
      
      sprintf(snapshotFile, "snapshot%d-%d-%d.ppm", 
	      NumRecReceived, now->tm_hour, now->tm_min);
      
      ppm1 = rawrec1DCP->ppm1; ppm2 = rawrec1DCP->ppm2;
      width = rawrec1DCP->width; height = rawrec1DCP->height;
      tmpBuff = rawrec1DCP->buff; 
      size = rawrec1DCP->size;
      tag = rawrec1DCP->tag;
    } else if (!strcmp(format, "Raw_data2DC\0")) {
      rawrec2DCP = (Raw_data2DC_ptr) event;
      if (verbose) fprintf(stdout, "Got event %d\n", rawrec2DCP->tag);
      
      sprintf(snapshotFile, "snapshot%d-%d-%d.ppm", 
	      NumRecReceived, now->tm_hour, now->tm_min);
      
      ppm1 = rawrec2DCP->ppm1; ppm2 = rawrec2DCP->ppm2;
      width = rawrec2DCP->width; height = rawrec2DCP->height;
      tmpBuff = rawrec2DCP->buff; 
      size = rawrec2DCP->size;
      tag = rawrec2DCP->tag;
    } else if (!strcmp(format, "Raw_data3DC\0")) {
      rawrec3DCP = (Raw_data3DC_ptr) event;
      if (verbose) fprintf(stdout, "Got event %d\n", rawrec3DCP->tag);
      
      sprintf(snapshotFile, "snapshot%d-%d-%d.ppm", 
	      NumRecReceived, now->tm_hour, now->tm_min);
      
      ppm1 = rawrec3DCP->ppm1; ppm2 = rawrec3DCP->ppm2;
      width = rawrec3DCP->width; height = rawrec3DCP->height;
      tmpBuff = rawrec3DCP->buff; 
      size = rawrec3DCP->size;
      tag = rawrec3DCP->tag;
    } else if (!strcmp(format, "Raw_data4DC\0")) {
      rawrec4DCP = (Raw_data4DC_ptr) event;
      if (verbose) fprintf(stdout, "Got event %d\n", rawrec4DCP->tag);
      
      sprintf(snapshotFile, "snapshot%d-%d-%d.ppm", 
	      NumRecReceived, now->tm_hour, now->tm_min);
      
      ppm1 = rawrec4DCP->ppm1; ppm2 = rawrec4DCP->ppm2;
      width = rawrec4DCP->width; height = rawrec4DCP->height;
      tmpBuff = rawrec4DCP->buff; 
      size = rawrec4DCP->size;
      tag = rawrec4DCP->tag;
    } else if (!strcmp(format, "Raw_data1DG\0")) {
      rawrec1DGP = (Raw_data1DG_ptr) event;
      if (verbose) fprintf(stdout, "Got event %d\n", rawrec1DGP->tag);
      
      sprintf(snapshotFile, "snapshot%d-%d-%d.pgm", 
	      NumRecReceived, now->tm_hour, now->tm_min);
      
      ppm1 = rawrec1DGP->ppm1; ppm2 = rawrec1DGP->ppm2;
      width = rawrec1DGP->width; height = rawrec1DGP->height;
      tmpBuff = rawrec1DGP->buff; 
      size = rawrec1DGP->size;
      tag = rawrec1DGP->tag;
    } else if (!strcmp(format, "Raw_data2DG\0")) {
      rawrec2DGP = (Raw_data2DG_ptr) event;
      if (verbose) fprintf(stdout, "Got event %d\n", rawrec2DGP->tag);
      
      sprintf(snapshotFile, "snapshot%d-%d-%d.pgm", 
	      NumRecReceived, now->tm_hour, now->tm_min);
      
      ppm1 = rawrec2DGP->ppm1; ppm2 = rawrec2DGP->ppm2;
      width = rawrec2DGP->width; height = rawrec2DGP->height;
      tmpBuff = rawrec2DGP->buff; 
      size = rawrec2DGP->size;
      tag = rawrec2DGP->tag;
    } else if (!strcmp(format, "Raw_data3DG\0")) {
      rawrec3DGP = (Raw_data3DG_ptr) event;
      if (verbose) fprintf(stdout, "Got event %d\n", rawrec3DGP->tag);
      
      sprintf(snapshotFile, "snapshot%d-%d-%d.pgm", 
	      NumRecReceived, now->tm_hour, now->tm_min);
      
      ppm1 = rawrec3DGP->ppm1; ppm2 = rawrec3DGP->ppm2;
      width = rawrec3DGP->width; height = rawrec3DGP->height;
      tmpBuff = rawrec3DGP->buff; 
      size = rawrec3DGP->size;
      tag = rawrec3DGP->tag;
    } else if (!strcmp(format, "Raw_data4DG\0")) {
      rawrec4DGP = (Raw_data4DG_ptr) event;
      if (verbose) fprintf(stdout, "Got event %d\n", rawrec4DGP->tag);
      
      sprintf(snapshotFile, "snapshot%d-%d-%d.pgm", 
	      NumRecReceived, now->tm_hour, now->tm_min);
      
      ppm1 = rawrec4DGP->ppm1; ppm2 = rawrec4DGP->ppm2;
      width = rawrec4DGP->width; height = rawrec4DGP->height;
      tmpBuff = rawrec4DGP->buff; 
      size = rawrec4DGP->size;
      tag = rawrec4DGP->tag;
    }
  }

  sfd = open(snapshotFile, O_CREAT|O_WRONLY, S_IRWXU|S_IRGRP|S_IROTH);
  sprintf(header, "%c%c\n%d %d\n%d\n", ppm1, ppm2, width, height, 255);  
  write(sfd, header, sizeof(header));
  if (tmpBuff != NULL) write(sfd, tmpBuff, size);
  close(sfd);
  
  if (NumRecReceived == NumRec)
    getUsage(&stopCpu, &stopUCpu, &stopSCpu, &stopTime);
    
  return;
} /* end eventHandler */


/*----------------------------------------------------------------------------*/

int
main(argc, argv)
     int argc; 
     char* argv[];
{
  EControlContext ec;
  IOContext ioc;
  int optc;  
  char* chanID = NULL;
  char placement;
  char* strlt_path = NULL;
  Streamlet_T strlt;
  Params_rectangle imgRectangle;
  Params_factor imgFactor;
  char paramsName[30];
  ECdata_struct imgRectangle_spec;
  ECdata_struct imgFactor_spec;

  StreamletName = NULL;
  local_progname = argv[0];

  Styx_asDaemon(0);             /* not a daemon, actually do not use syslog */
  Styx_setDebugLevel(0);        /* no debugging unless requested */
  Styx_setProgname(argv[0]);

  /*vvv*/
  StreamletLoc = SLOC_REMOTE;
  /*StreamletLoc = SLOC_LOCAL;*/
  /*^^^*/

  while ((optc = getopt_long(argc, argv, "+vhmt:p:n:s:c:", 
			     longopts, (int*) 0)) != EOF) {
    switch (optc) {
    case 'v':
      verbose= TRUE;
      Styx_setDebugLevel(7);
      Styx_privateLog(0);
      break;
    case 'h':
      usage(stdout, 0);
      break;
    case 'm':
      MeasureMe = TRUE;
      break;
    case 'n':
      NumRec = atoi(optarg);
      break;
    case 'c':
      chanID = optarg;
      break;
    case 't':
      StreamletName = optarg;
      break;
    case 's':
      Freq = atoi(optarg);
      if (Freq < 0) usage(stderr, 1);
      break;
    case 'p':
      placement = optarg[0];
      if (placement == 'r') 
	StreamletLoc = SLOC_REMOTE;
      else if (placement != 'l') 
	StreamletLoc = SLOC_LOCAL;
      else  usage(stderr, 1);
      break;
    default:
      usage(stderr, 1);
    }
  }

  if (StreamletName == NULL && StreamletLoc == 0) usage (stderr, 1);
  if (optind > argc) usage(stderr, 1); /* no extra args */

  if (verbose) {
    fprintf(stdout, "From channel:        \t%s\n", chanID);
    fprintf(stdout, "Number of frames:    \t%d\n", NumRec);
    fprintf(stdout, "Streamlet name:      \t%s\n", StreamletName);
    fprintf(stdout, "Streamlet placement: \t%d\n", StreamletLoc);
  }
  
  cm = CManager_create();
  ec = ECho_CM_init(cm);
  ioc = create_IOcontext();	/* Context to keep formats */
  
  /* Register formats with ioc */
  avs_registerFormats(ioc);

  if (signal(SIGINT, interruptHandler) == SIG_ERR) 
    Styx_errQuit("Signal error");

  if (StreamletLoc == SLOC_REMOTE) {
    if (verbose) fprintf(stdout, "Deriving new transformed channel\n");
    strlt = Strlt_new();

    strlt_path = getenv ("AVS_STRLT_PATH");
    if (strlt_path == NULL) {
      if (verbose) fprintf (stdout, "AVS_STRLT_PATH not set, using built-in default\n");
      strlt_path = default_strlt_path;
    }
    if (verbose) fprintf (stdout, "Streamlet path is %s\n", strlt_path);
      
    if (Strlt_init(strlt, strlt_path, StreamletName) != 0)
      Styx_errQuit("Failed getting streamlet %s\n", StreamletName); 
    
    if (verbose) {
      fprintf(stdout, "Streamlet name          %s<<<\n", 
	      Strlt_getName(strlt));
      fprintf(stdout, "Streamlet input         %s<<<\n", 
	      Strlt_getInputChan(strlt)); 
      fprintf(stdout, "Streamlet output        %s<<<\n", 
	      Strlt_getOutputChan(strlt));
      fprintf(stdout, "Streamlet takes params? %c<<<\n", 
	      Strlt_takesParams(strlt)? 'Y':'N');
      if (Strlt_takesParams(strlt))
	fprintf(stdout, "Streamlet params        %s<<<\n", 
		Strlt_getParams(strlt));  
      fprintf(stdout, "Streamlet code\n%s<<<\n", Strlt_getCode(strlt)); 
    }

    /* Prepare parameters  */
    /* This is ugly but I cannot think of a better way, perhaps argv? */
    imgRectangle.x = 0;
    imgRectangle.y = 0;
    imgRectangle.width = 160;
    imgRectangle.height = 120;
    strcpy(paramsName, "Params_rectangle\0");
    imgRectangle_spec.data_field_list = 
      avs_getFormatFieldList(ioc, paramsName);
    imgRectangle_spec.data_subformat_list = NULL;
    imgRectangle_spec.initial_value = &imgRectangle;

    imgFactor.factor = 2; /* Here we are always halving/doubling the size of imgs */
    strcpy(paramsName, "Params_factor\0");
    imgFactor_spec.data_field_list = 
      avs_getFormatFieldList(ioc, paramsName);
    imgFactor_spec.data_subformat_list = NULL;
    imgFactor_spec.initial_value = &imgFactor;


    if (!strcmp(Strlt_getName(strlt), "cropImage.ecl")) {
      if (verbose) fprintf(stdout, "Crop image ... deriving channel ...\n");
      derivedChan = 
	EChannel_typed_derive_data(ec, chanID, Strlt_getCode(strlt), 
				   avs_getFormatFieldList(ioc, Strlt_getInputChan(strlt)), 
				   NULL, &imgRectangle_spec);
    } else if (!strcmp(Strlt_getName(strlt), "halfRate.ecl")) {      
      NumRec = NumRec/2;	/* Adjust expected number of records */
      if (verbose) fprintf(stdout, "Half rate ... deriving channel ...\n");
      derivedChan = EChannel_derive (ec, chanID, Strlt_getCode(strlt));
      if (NULL==derivedChan) fprintf(stderr," [HalfRate CHanId failed\n");
    } else if (!strcmp(Strlt_getName(strlt), "reduceImage.ecl")) {
      if (verbose) fprintf(stdout, "Half image ... deriving channel ...\n");
      derivedChan = 
	EChannel_typed_derive_data (ec, chanID, Strlt_getCode(strlt), 
				    avs_getFormatFieldList(ioc, Strlt_getInputChan(strlt)), 
				    NULL, &imgFactor_spec);
    } else if (!strcmp(Strlt_getName(strlt), "greyImage.ecl")) {
      if (verbose) fprintf(stdout, "Grey rate ... deriving channel ...\n");
      derivedChan = 
	EChannel_typed_derive_data (ec, chanID, Strlt_getCode(strlt), 
				    avs_getFormatFieldList(ioc, Strlt_getInputChan(strlt)), 
				    NULL, NULL);
    } else {
      fprintf(stderr, "Streamlet unrecognized\n");
      if (StreamletLoc == SLOC_REMOTE) EChannel_destroy(derivedChan);
      else EChannel_destroy(chan);
      CManager_close(cm);
      exit(1);
    }
    
    if(verbose) fprintf(stdout, "New channel derived (output: %s)\n", 
			Strlt_getOutputChan(strlt));
    
    if (verbose) fprintf(stdout, "Subscribing to new channel\n");
    
    if (MeasureMe) {
      sensNet_Init(Freq);
      sensNet_SetHistLen(NumRecReceived/Freq, SENSNET_READ);
    }

    getUsage(&startCpu, &startUCpu, &startSCpu, &startTime);
    
    ECsink_typed_subscribe(derivedChan, 
			   avs_getFormatFieldList(ioc, Strlt_getOutputChan(strlt)), 
			   NULL, eventHandler, 
			   Strlt_getOutputChan(strlt));
    
  } else {			/* StreamletLoc == SLOC_LOCAL */
    if (verbose) fprintf(stdout, "Opening given channel %s\n", chanID);
    chan = EChannel_open(ec, chanID); 
    if (chan == NULL) Styx_errQuit("Failed to open channel\n");
    
    if (verbose) fprintf(stdout, "Subscribing to new channel\n");
    
    if (MeasureMe) {
      sensNet_Init(Freq);
      sensNet_SetHistLen(NumRecReceived/Freq, SENSNET_READ);
    }

    getUsage(&startCpu, &startUCpu, &startSCpu, &startTime);
    ECsink_typed_subscribe(chan, Raw_data_fld, NULL, eventHandler, NULL);
  }
  
  CMsleep(cm, 6000);		/* 10' and go away  */

  dumpStats();  
  if (StreamletLoc == SLOC_REMOTE) EChannel_destroy(derivedChan);
  else EChannel_destroy(chan);
  CManager_close(cm);

  exit(EXIT_SUCCESS);
} /* main */
