/* -*-C-*-
*******************************************************************************
*
* File:         ebw_sink.c
* RCS:          $Id: ebw_sink.c,v 1.1.1.1 2003/06/11 18:48:52 wyan Exp $
* Description:  ECho bandwidth experiment
* Author:       Fabian E. Bustamante
*               Systems Research Group
*               College of Computing
*               Georgia Institute of Technology
* Created:      Wed Sep 19, 2001 at 23:00:54
* Modified:     Mon Oct 01, 2001 at 08:11:26 fabianb@cc.gatech.edu
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
* (C) Copyright 2001, Georgia Institute of Technology, all rights reserved.
*
*******************************************************************************
*/

#if HAVE_CONFIG_H
#  include <src/config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include <getopt.h>
#include <io.h>
#include <atl.h>
#include <cm.h>
#include <echo.h>

#include <ebw_common.h>
#include <stats.h>
#include <sensNet.h>

static CManager cm;
static EChannel chan;
static int verbose = 0;
static int Measure = 0;
static int TotalEvents;
static int NumEvents = 1;
static int Freq;
static char *local_progname;

static struct option longopts[] =
{
  {"verbose", no_argument, NULL, 'v'},
  {"help", no_argument, NULL, 'h'},
  {"measure", no_argument, NULL, 'm'},
  {"events", required_argument, NULL, 'e'},
  {"frequency", required_argument, NULL, 'f'},
  {"channel", required_argument, NULL, 'c'},
  {NULL, no_argument, NULL, 0}
};

/*------------------------------------------------------------------------------*/

/*
 * dump_stats:
 *
 * Parameters: 
 * 
 * Return: 
 *
 */

void 
dump_stats() 
{
  int i;
  double *bwHist;
  
  if (Measure) {
    bwHist = (double*) malloc(((int) (TotalEvents/Freq)) * sizeof(double));
    if (bwHist == NULL) {
      perror("Not enough resources");
      exit(1);
    }
    
    fflush(stdout);
    sensNet_GetHistory(bwHist, (int) (TotalEvents/Freq), SENSNET_READ);
    for (i = 0; i < (int) (TotalEvents/Freq); i++) 
      fprintf(stdout, "Bw %.3f Mbps\n", bwHist[i]);

    fprintf(stdout, "Bw mean %.3f Mbps (sdev %.3f)\n", 
	    Stats_mean (bwHist, (int) (TotalEvents/Freq)),
	    Stats_sdev (bwHist, (int) (TotalEvents/Freq)));

    fflush(stdout);
    sensNet_Finish(); 
    
    free(bwHist);
  }

  EChannel_destroy(chan);
  CManager_close(cm);

  exit(0);
} /* end dump_stats */

/*----------------------------------------------------------------------------*/

/*
 * interruptHandler: To go out nicely
 *
 * Parameters: signal number
 * 
 * Return: -
 *
 */

static void
interruptHandler(signo)
     int signo;
{
  if (Measure)
    sensNet_Finish();
    
  EChannel_destroy(chan);
  CManager_close(cm);

  exit(0);
} /* end interruptHandler */


/*------------------------------------------------------------------------------*/

/*
 * handler:
 *
 * Parameters: 
 * 
 * Return: 
 *
 */

void 
handler (event, client_data, attrs)
     void* event;
     void* client_data;
     attr_list attrs;
{
  ++NumEvents;
  /* fprintf(stdout, "Event %d/%d\n", NumEvents, TotalEvents); */
  if (NumEvents == TotalEvents)
    dump_stats();

  return;
} /* end handler */

/*----------------------------------------------------------------------------*/

/*
 * usage: Prints out 'usage' string
 *
 * Parameters: target 'stream' and 'status' to quit with
 *
 * Returns: -
 *
 */

static void 
usage (stream, status)
     FILE* stream; 
     int status;
{
  fprintf(stream, "Usage: %s [-vhm] -c  channel-id -n number-frames\n"
	  "-v --verbose \tverbose mode\n"
	  "-h --help    \tgive this help\n"
	  "-m --measure \tmeasure this end\n"
	  "-e --events  \tnumber of events expected (0 == infinite)\n"
	  "-f --frequency            \tsampling frequency (>= 1)\n"
	  "-c --channel \tchannel id\n"
	  , local_progname);
  exit(status);
} /* end usage */

/*------------------------------------------------------------------------------*/

int 
main(argc, argv)
     int argc; 
     char* argv[];
{
  EControlContext ec;
  int optc;
  char* chanID = NULL;

  local_progname = argv[0];
  if (argc < 7) usage(stderr, 1);
  while ((optc = getopt_long(argc, argv, "+vhmc:e:f:", 
			     longopts, (int*) 0)) != EOF) {
    switch (optc) {
    case 'v':
      verbose = 1;
      break;
    case 'h':
      usage(stdout, 0);
      break;
    case 'm':
      Measure = 1;
      break;
    case 'c':
      chanID = optarg;
      break;
    case 'e':
      TotalEvents = atoi(optarg);
      if (TotalEvents < 0) usage(stderr, 1);
      break;
    case 'f':
      Freq = atoi(optarg);
      if (Freq < 1) usage(stderr, 1);
      break;
    default:
      usage(stderr, 1);
    }
  }
  
  /* validate arguments */
  if ((TotalEvents < 1) || Freq < 1) usage(stderr, 1);

  if (optind > argc) usage(stderr, 1); /* no extra args */
  
  if (signal(SIGINT, interruptHandler) == SIG_ERR) {
    perror("Signal error");
    exit(1);
  }

  if (verbose) fprintf(stdout, "From channel: \t%s\n", chanID);

  cm = CManager_create();
  ec = ECho_CM_init(cm);

  if (verbose) fprintf(stdout, "Opening channel\n");
  chan = EChannel_open(ec, chanID); 
  if (chan == NULL) {
    perror("Failed to open channel\n");
    exit(1);
  } else
    if (verbose) fprintf(stdout, "Channel open\n");

  if (Measure) {
    sensNet_Init(Freq);
    sensNet_SetHistLen(TotalEvents/Freq, SENSNET_READ);
  }

  if (verbose) fprintf(stdout, "Subscribing to new channel\n");
  ECsink_typed_subscribe(chan, Datarec_fld, NULL, handler, NULL);

  while (1) {
    CMpoll_network(cm);
  }

  exit(0);
} /* end main */
