/* -*-C-*-
*******************************************************************************
*
* File:         sensNet.c
* RCS:          $Id: sensNet.c,v 1.1.1.1 2003/06/11 18:48:52 wyan Exp $
* Description:  Sensor for network effective throughput
*               Notice that we are timing in between "attempts" to read/write,
*               under the assumption you would not do much in between.
*
*                        >   what we timed  <                 
*               Time  ---+---+---+---+---+---+---+---+---+---+---+---+---+-- 
*                        ^   ^               ^   ^
*                        |   |               |   |
*                        |   Stop read       |   Stop read
*                        Start read (n)      Start read (n+1)
*
* Author:       Fabian E. Bustamante
*               Systems Research Group
*               College of Computing
*               Georgia Institute of Technology
* Created:      Tue Sep 18, 2001 at 21:31:50
* Modified:     Thu Oct 11, 2001 at 23:10:30 fabianb@cc.gatech.edu
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
* (C) Copyright 2001, Georgia Institute of Technology, all rights reserved.
*
*******************************************************************************
*/

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sensNet.h>
#include <sys/time.h>

#include <tv.h>

#include <io.h>
#include <atl.h>
#include <cm.h>
#include <echo.h>
 
extern int (*cm_write_hook)(int); /* cm hook into write */
extern int (*cm_preread_hook)(int, char*); /* cm hook into read */

int write_handler PARAMS((int write_length));
int read_handler PARAMS((int read_length, char* str));

int HistLen[2] = {-1, -1};
int LastHistEntry[2] = {-1, -1};
double *History[2];		/* Iliffe vector with history of 
				   last HistLen measurements */

int Frequency;			/* read/write in between samples */
struct timeval LastWrite_tv, LastRead_tv; /* time of last read/write */
int CountWrite, CountRead;	/* num of read/write to enforce freq */
int AccWrite, AccRead;		/* read/write accumulators */

/*----------------------------------------------------------------------------*/

/* 
 * sensNet_Init: Initializes network sensor
 * 
 * Parameters: frequency of sampling as number of reads/writes in between.
 *   Every read/write we increment the corresponding accumulator. Every 
 *   'frequency' read/write we compute bandwith and entry it in history.
 *
 * Returns: Zero on success, or -1 if an error occurred.
 *
 */

int
sensNet_Init(frequency)
     int frequency; 
{
  if (frequency < 1) return -1;
  else Frequency = frequency;

  CountWrite = 0;
  AccWrite = 0;
  cm_write_hook = write_handler;
  tv_zero(&LastWrite_tv);

  CountRead = 0;
  AccRead = 0;
  cm_preread_hook = read_handler;
  tv_zero(&LastRead_tv);

  return 0;
} /* end sensNet_Init */


/*----------------------------------------------------------------------------*/

/* 
 * sensNet_Stop: Stop channels monitoring
 * 
 * Parameters: Op flag (read or write)
 *
 * Returns: Zero on success, or -1 if an error occurred.
 *
 */

int
sensNet_Stop(op)
     int op; 
{
  if (op != SENSNET_READ && op != SENSNET_WRITE) return -1;
  if (op == SENSNET_WRITE) cm_write_hook = NULL;
  else cm_preread_hook = NULL;

  return 0;
} /* end sensNet_Stop */

/*----------------------------------------------------------------------------*/

/* 
 * sensNet_Continue: Continue channels monitoring; This re-starts timer, thus 
 *    it could translate in one artificially high bandwidth
 * 
 * Parameters: Op flag (read or write)
 *
 * Returns: Zero on success, or -1 if an error occurred.
 *
 */

int
sensNet_Continue(op)
     int op; 
{
  struct timeval tv;

  if (op != SENSNET_READ && op != SENSNET_WRITE) return -1;
  if (gettimeofday(&tv, NULL) == -1) {
    perror("Cannot get timestamp");
    return -1;
  }
  
  if (op == SENSNET_WRITE) {
    cm_write_hook = write_handler;
    tv_assign(&LastWrite_tv, &tv);
  } else {
    cm_preread_hook = read_handler;
    tv_assign(&LastRead_tv, &tv);
  }

  return 0;
} /* end sensNet_Continue */


/*----------------------------------------------------------------------------*/

/* 
 * sensNet_Finish: Clean up after us
 * 
 * Parameters: -
 *
 * Returns: Zero on success, or -1 if an error occurred.
 *
 */

int 
sensNet_Finish() 
{
  if ((HistLen[SENSNET_READ] != -1) || (HistLen[SENSNET_WRITE] != -1))
    sensNet_FreeHistory();
  return 0;
} /* end sensNet_Finish */


/*----------------------------------------------------------------------------*/

/* 
 * sensNet_SetHistLen: Set the length of history we should remember, history
 *    is seen as circular
 * 
 * Parameters: histLen - length of history to keep and op flag
 *
 * Returns: old histLen or 0 if there was none, -1 in error 
 *
 */

int 
sensNet_SetHistLen(newHistLen, op)
     int newHistLen; 
     int op; 
{
  int oldHistLen;
  double *newHistory;
  int i, j;

  /* dealing with the obvious */
  if (op != SENSNET_READ && op != SENSNET_WRITE) return -1;
  if (newHistLen < 0) return -1;
  if (newHistLen == HistLen[op])
    return HistLen[op];

  /* Get space for new history */
  newHistory = (double*) malloc(newHistLen*sizeof(double));
  if (newHistory == NULL) {
    perror("Not enough resources");
    return -1;
  }

  /* Now, do we need to save past history ? */
  if (HistLen[op] == -1) {
    oldHistLen = 0;
    LastHistEntry[op] = -1;
  } else {
    oldHistLen = HistLen[op];

    if (LastHistEntry[op] != -1) {
      /* Have some actual history to save */
      for (i = LastHistEntry[op], j = newHistLen - 1;
	   i != LastHistEntry[op] + 1 && j >= 0; j--) {

	newHistory[j] = History[op][i];
	if (i == 0) i = HistLen[op] - 1; 
	else --i;
      }
      LastHistEntry[op] = HistLen[op] - 1;
    }

    free(History[op]);
  }

  HistLen[op] = newHistLen;
  History[op] = newHistory;

  return oldHistLen;
} /* end sensNet_SetHistLen */


/*----------------------------------------------------------------------------*/

/* 
 * sensNet_GetHistory: Get what we have so far in history
 * 
 * Parameters: Place holder (must be pre-alloc), its size, and op flag
 *
 * Returns: Zero on success, or -1 if an error occurred.
 *
 */

int 
sensNet_GetHistory(history, len, op)
     double history[]; 
     int len; 
     int op; 
{
  int i;
  int trueLen;

  if (op != SENSNET_READ && op != SENSNET_WRITE) 
    return -1;
  if (history == NULL) return -1;

  if (len > HistLen[op])
    trueLen = HistLen[op];
  else 
    trueLen = len;

  for (i = 0; i < trueLen; i++)
    history[i] = History[op][i];
  
  return trueLen;
} /* end sensNet_GetHistory */


/*----------------------------------------------------------------------------*/

/* 
 * sensNet_GetLast: Get last entry in history
 * 
 * Parameters: op flag
 *
 * Returns: Last bandwidth registered in history, or -1 if an error occurred.
 *
 */

double
sensNet_GetLast(op) 
     int op;
{
  if (op != SENSNET_READ && op != SENSNET_WRITE) 
    return -1;

  if (History[op] != NULL) 
    return History[op][LastHistEntry[op]];

  return -1;
} /* end sensNet_GetLast */


/*----------------------------------------------------------------------------*/

/* 
 * sensNet_FreeHistory: Free accumulated history (both, read and write)
 * 
 * Parameters: -
 *
 * Returns: 0 if everything is OK
 *
 */

int 
sensNet_FreeHistory()
{
  if (HistLen[SENSNET_READ] != -1) {
    free(History[SENSNET_READ]);
    HistLen[SENSNET_READ] = -1;
    History[SENSNET_READ] = NULL;
    LastHistEntry[SENSNET_READ] = -1;
  }

  if (HistLen[SENSNET_WRITE] != -1) {
    free(History[SENSNET_WRITE]);
    HistLen[SENSNET_WRITE] = -1;
    History[SENSNET_WRITE] = NULL;
    LastHistEntry[SENSNET_WRITE] = -1;
  }

  return 0;
} /* end sensNet_FreeHistory */


/*----------------------------------------------------------------------------*/

/* 
 * write_handler: Called before the write is done, we timed in between
 *    (see note on header)
 * 
 * Parameters: amount to write out
 *
 * Returns: 1 for the operation to be done
 *
 */

int
write_handler(write_length)
     int write_length;
{
  struct timeval tv, periodtv;
  double bandwidth;
  u_long usecs;
  static int firstWrite = 1;

  AccWrite += write_length;
  ++CountWrite;

  if (firstWrite) {
    firstWrite = 0;
    if (gettimeofday(&tv, NULL) == -1) {
      perror("Cannot get timestamp");
      return -1;
    }
    tv_assign(&LastWrite_tv, &tv);
  }

  if (CountWrite == Frequency) {
    if (gettimeofday(&tv, NULL) == -1) {
      perror("Cannot get timestamp");
      return -1;
    }
    tv_minus2(&periodtv, &tv, &LastWrite_tv);

    usecs = tv_toUsecs(&periodtv);
    if (usecs > 0)
      bandwidth = ((double) (AccWrite * 8)) / usecs;
    else 
      perror("usecs 0");

    if (LastHistEntry[SENSNET_WRITE] == HistLen[SENSNET_WRITE]) 
      LastHistEntry[SENSNET_WRITE] = 0;
    else ++LastHistEntry[SENSNET_WRITE];
    History[SENSNET_WRITE][LastHistEntry[SENSNET_WRITE]] = bandwidth;

    AccWrite = 0;
    CountWrite = 0;
    tv_assign(&LastWrite_tv, &tv);
  }

  return 1;
} /* end write_handler */


/*----------------------------------------------------------------------------*/

/* 
 * read_handler: Called before the full read is done, we timed in between
 *    (see note on header)
 *
 * Parameters: -
 *
 * Returns: 1 for the operation to be done
 *
 */

int
read_handler(read_length, str)
     int read_length; 
     char *str;
{
  struct timeval tv, periodtv;
  double bandwidth;
  u_long usecs;
  static int firstRead = 1;
  static int headerRead = 0;

  AccRead += read_length;

  if (firstRead) {
    firstRead = 0;
    if (gettimeofday(&tv, NULL) == -1) {
      perror("Cannot get timestamp");
      return -1;
    }
    tv_assign(&LastRead_tv, &tv);
  }

  /* This a dangerous hack - CM read 2 times 4 bytes from the header, I want to
     count the whole thing (3 reads) once */
  if (headerRead < SENSNET_CM_HEADER_READS) {
    ++headerRead;
    return 1;
  } else {
    headerRead = 0;

    ++CountRead;
    if (CountRead == Frequency) {
      if (gettimeofday(&tv, NULL) == -1) {
	perror("Cannot get timestamp");
	return -1;
      }
      tv_minus2(&periodtv, &tv, &LastRead_tv);
      
      usecs = tv_toUsecs(&periodtv);
      if (usecs > 0) bandwidth = ((double) (AccRead * 8)) / usecs;
      
      if (LastHistEntry[SENSNET_READ] == HistLen[SENSNET_READ]) 
	LastHistEntry[SENSNET_READ] = 0;
      else ++LastHistEntry[SENSNET_READ];
      History[SENSNET_READ][LastHistEntry[SENSNET_READ]] = bandwidth;
      
      AccRead = 0;
      CountRead = 0;
      tv_assign(&LastRead_tv, &tv);
    }
  }

  return 1;
} /* end read_handler */
