/* -*-C-*-
*******************************************************************************
*
* File:         argl.c
* RCS:          $Id: argl.c,v 1.1.1.1 2003/06/11 18:49:00 wyan Exp $
* Description:  
* Author:       Fabian E. Bustamante
*               Systems Research Group
*               College of Computing
*               Georgia Institute of Technology
* Created:      Wed Jul 04, 2001 at 22:07:19
* Modified:     Tue Oct 02, 2001 at 23:36:47 fabianb@cc.gatech.edu
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
* (C) Copyright 2001, Georgia Institute of Technology, all rights reserved.
*
*******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "argl.h"

#define UNSPLIT do { if (val != NULL) s[eqpos] = '='; } while (0)

typedef struct {
    char * name;
    unsigned long varsize;    
    Argl_item_T type;
    void * var;
} Argl_item;

int Argl_errno;

static char *yesstrings[] = {"yes", "YES", "Yes", "Y", "y", 
			     "1", "true", "on", "enable", NULL};
static char *nostrings[] = {"no", "NO", "No", "N", "n", 
			    "0", "false", "off", "disable", NULL};

static Argl_item ai[ARGL_MAX_ITEMS];
static int num_argl_items = 0;

static char *errstr[] = {
    "INT/BOOL args must be \"int\" vars",
    "Bad format of argument"
};

/*------------------------------------------------------------------------------*/

/*
 * Argl_strerror:
 *
 * Parameters:
 *
 * Returns: A short string giving a plain-text description for an error code.
 *
 */

char * 
Argl_strerror (int errno)
{
  return errstr[errno-1];
} /* end Argl_strerror */


/*------------------------------------------------------------------------------*/

/*
 * Argl_reset: Resets internal state and gets ready for new parsing
 *
 * Parameters:
 *
 * Returns: Zero for success, -1 for error
 *
 */

int
Argl_reset ()
{
  num_argl_items = 0;
  return 0;
} /* end Argl_reset */


/*------------------------------------------------------------------------------*/

/*
 * Argl_register: Registers a particular command line parameter.  The
 *   command line syntax is <name>=<value> (no dashes).  "varsize" 
 *   specifies the size of "var", important for char[] etc.  "ait" is 
 *   the type and style (see "Argl_item_T").
 *
 * Parameters:
 *
 * Returns: Zero for success, -1 for error
 *
 */

int
Argl_register (char *name, void *var, unsigned long varsize, Argl_item_T ait)
{
  Argl_item * aip = ai+(num_argl_items++);
  aip->name = name;
  aip->type = ait;
  aip->var = var;
  aip->varsize = varsize;

  if ((ait == ARGL_INT || ait == ARGL_BOOL) && (varsize != sizeof(int))) {
    Argl_errno = ARGL_ERRNO_VARSIZE;
    return -1;
  }
  return 0;
} /* end Argl_register */


/*------------------------------------------------------------------------------*/

/*
 * Argl_parse_1:
 *
 * Parameters:
 *
 * Returns: 1 if recognized, 0 if not recognized, -1 on error.  If 0,
 *          string is not changed, otherwise it will be changed.
 *
 */

int
Argl_parse_1 (char *s)
{
  char * val;
  int eqpos, i, j;
  char *endptr;

  fprintf (stderr, "string=(%s)\n", s);
  /* Find the = in the argument (if any) and split the string into two.
     A pointer to the argument value is saved in "val". */
  val = 0;
  for (eqpos = 0; s[eqpos] != 0; eqpos++)
    if (s[eqpos] == '=') {
      val = s+eqpos+1;
      s[eqpos] = 0;
      break;
    }    

  /* Look whether we have a match. */
  if (strcmp (s, "Argl_read") == 0) {
    int res;
    FILE * f = fopen (val, "r");
    if (f == NULL) {
      perror ("In Argl_parse_1: fopen failed");
      return -1;
    }
    res = Argl_parse_file (f);
    if (res > 0) {
      fprintf (stderr, "In Argl_parse_1: parse error (file %s line %d)\n", val, res);
      return -1;
    }
  } else
    for (i = 0; i < num_argl_items; i++)
      if (strcmp (s, ai[i].name) == 0) {
	switch (ai[i].type) {
	case ARGL_INT:
	  if (val == NULL) goto error;
	    val[strlen(val)-1] = '\0';
	  *((int *)ai[i].var) = strtol (val, & endptr, 0);
	  if (endptr == val) goto error;
	  if (*endptr != 0) { /* there might be K,k,M or G at the end. */
	    switch (*endptr) {			    
	    case 'K':
	    case 'k':
	      *((int*)ai[i].var) <<= 10;
	      break;
	    case 'M':
	      *((int*)ai[i].var) <<= 20;
	      break;
	    case 'G':
	      *((int *)ai[i].var) <<= 30;
	      break;
	    default:
	      goto error;
	    }
	    endptr++;
	    if (*endptr != 0) goto error;
	  }
	  goto recognized;
	case ARGL_DOUBLE:
	  if (val == NULL) goto error;
	  val[strlen(val)-1] = '\0';
	  *((double *)(ai[i].var)) = strtod (val, & endptr);
	  if (endptr == val) goto error;
	  if (*endptr != 0) goto error; // no trailing stuff permitted
	  goto recognized;
	case ARGL_STRING:
	  if (val == NULL) goto error; 
	  val[strlen(val)-1] = '\0';
	  *((char **)ai[i].var) = strdup(val);
	  goto recognized;
	case ARGL_BOOL:
	  if (val == NULL) {
	    *((int*)ai[i].var) = 1;
	    goto recognized;
	  } else {
	    /* char **str; */
	    val[strlen(val)-1] = '\0';
	    for (j = 0; yesstrings[j] != NULL; ++j) {
	      if (strcmp(yesstrings[j], val) == 0) {
		*((int *)ai[i].var) = 1;
		goto recognized;
	      }
	    }
	    for (j = 0; nostrings[j] != NULL; ++j) {
              fprintf (stderr, "val-%s\n", val);
              fprintf (stderr, "test-%s\n", nostrings[j]);
	      if (strcmp(nostrings[j], val) == 0) {
		*((int *)ai[i].var) = 0;
                fprintf (stderr, "it=%d, name=(%s)\n", *((int*)ai[i].var), ((char*)ai[i].name));
                
		goto recognized;
	      }
	    }
	    goto error;
	  }
	} /* switch */
      } /* if */
  /* not_recognized:   not used */
  return 0;
 recognized:
  return 1;
 error:
  Argl_errno = ARGL_ERRNO_ARG_FORMAT;
  return -1;
} /* end Argl_parse_1 */


/*------------------------------------------------------------------------------*/

/*
 * Argl_parse: If parsing proceeds till the end of argv, argc is
 *   returned.  argc is not changed.  Only those elements of argv that
 *   are recognized are changed.  If there is a parse error, -(errpos)
 *   is returned.  errpos is the index of the erroneous option in in
 *   argv[].
 *
 * Parameters:
 *
 * Returns: The position of the first unrecognized argument.
 *
 */

int
argl_parse (int argc, char **argv)
{
  int pos;

  pos = 1;
  while (pos < argc) {
    int res = Argl_parse_1 (argv[pos]);
    if (res == 1) {
      int i;
      argc--;
      for (i = pos; i < argc; i++)
	argv[i] = argv[i+1];
    }
    else 
      if (res < 0) return res;
    else pos++;
  }
  return argc;
} /* end Argl_parse */


/*------------------------------------------------------------------------------*/

/*
 * Argl_parse_file: Reads command line options from a file. One option
 *    per line.  Blanks at the beginning of a line are ignored. Entirely
 *    blank lines or lines with # as first non-blank are ignored. Lines
 *    must not be longer than ARGL_MAX_LINE_WIDTH.  If there are no
 *    errors, 0 is returned, otherwise the number of the erroneous line
 *    (first line of file is line 1).
 *
 * Parameters:
 *
 * Returns:
 *
 */

int
Argl_parse_file (FILE * f)
{
  char line[ARGL_MAX_LINE_WIDTH];
  char * start;
  int res, lineno;

  lineno = 1;
  while (!feof(f)) {
    memset (line, 0, ARGL_MAX_LINE_WIDTH);
    fgets (line, ARGL_MAX_LINE_WIDTH-1, f);
    start = line;
    while (*start == ' ' || *start == '\t')
      start++;
    if (*start == '\n' || *start == '#')
      continue;
    fprintf (stderr, "start is (%c)\n", *start);
    res = Argl_parse_1 (start);
    if (!res) 
      return lineno;
    /* ^^^ there must not be any non-argl lines in parsed files.
     * Support for that will be added later (and has to be specified
     * first anyways).
     */
    
    lineno++;
  }
  return 0;
} /* end Argl_parse_file */

