/* -*-C-*-
*******************************************************************************
*
* File:         avs_sink.c
* RCS:          $Id: avs_sink.c,v 1.1.1.1 2003/06/11 18:49:00 wyan Exp $
* Description:  AVS Sink
* Author:       Fabian Ernesto Bustamante
*               Systems Research Group
*               College of Computing
*               Georgia Institute of Technology
* Created:      Fri Feb 16, 2001 at 00:06:48
* Modified:     Fri Oct 12, 2001 at 09:58:27 fabianb@cc.gatech.edu
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
* (C) Copyright 2001, Georgia Institute of Technology, all rights reserved.
*
*******************************************************************************
*/

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/types.h>  
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>

#include <io.h>
#include <atl.h>
#include <cm.h>
#include <echo.h>

#include <getopt.h>
#include <hrtime.h>
#include <tv.h>
#include <avs_util.h>
#include <avs_raw.h>
#include <styx.h>
#include <sensNet.h>
#include <stats.h>

static char* local_progname;
static boolean verbose = FALSE;		/* verbose mode */
static boolean MeasureMe = FALSE;
static int Freq = 2;
static CManager cm;
static EChannel chan; 
static struct timeval startTime, stopTime;
static struct timeval startCpu, stopCpu;
static struct timeval startUCpu, stopUCpu;
static struct timeval startSCpu, stopSCpu;
static int NumRec = 0;
static int TotalNumRec;

/* Command line arguments */
static struct option longopts[] =
{
  {"verbose", no_argument, NULL, 'v'},
  {"help", no_argument, NULL, 'h'},
  {"measure", no_argument, NULL, 'm'},
  {"sample", required_argument, NULL, 's'},
  {"numframes", required_argument, NULL, 'n'},
  {"channel", required_argument, NULL, 'c'},
  {NULL, no_argument, NULL, 0}
};


/*----------------------------------------------------------------------------*/

/*
 * usage: Prints out 'usage' string
 *
 * Parameters: target 'stream' and 'status' to quit with
 *
 * Returns: -
 *
 */

static void 
usage (FILE* stream, int status)
{
  fprintf(stream, "Usage: %s [-vhm] -c channel-id -n number-frames -s freq\n"
	  "-v --verbose \tverbose mode\n"
	  "-h --help    \tgive this help\n"
	  "-m --measure \tset bandwidth measuring at client\n"
	  "-s --sample  \tsampling frequency for measurements\n"
	  "-n --numframes \tnumber of frames expected (0 == infinite)\n"
	  "-c --channel \tchannel id\n"
	  , local_progname);
  exit(status);
} /* end usage */


/*----------------------------------------------------------------------------*/

/*
 * dumpStats:
 *
 * Parameters: -
 *
 * Returns: -
 *
 */

void
dumpStats()
{
  struct timeval usedCpu, usedUCpu, usedSCpu, usedTime;
  double *bwHist;
      
  tv_minus2(&usedTime, &stopTime, &startTime);
  tv_minus2(&usedCpu, &stopCpu, &startCpu);
  tv_minus2(&usedUCpu, &stopUCpu, &startUCpu);
  tv_minus2(&usedSCpu, &stopSCpu, &startSCpu);
  fprintf(stdout, "\n----------------------------------------\n");
  fprintf(stdout, "Sink performance data\n");
  fprintf(stdout, "---------------------\n");
  fprintf(stdout, "Total time:       \t%ld ms\n", tv_toMsecs(&usedTime)); 
  fprintf(stdout, "Time per event:   \t%.3g ms\n", 
	  (float) tv_toMsecs(&usedTime)/NumRec);
  fprintf(stdout, "CPU usage:        \t%ld ms (%.3g) \n", 
	  tv_toMsecs(&usedCpu), 
	  (float) tv_toMsecs(&usedCpu)/tv_toMsecs(&usedTime));
  fprintf(stdout, "User CPU usage:   \t%ld ms (%.3g)\n", 
	  tv_toMsecs(&usedUCpu), 
	  (float) tv_toMsecs(&usedUCpu)/tv_toMsecs(&usedTime));
  fprintf(stdout, "System CPU usage: \t%ld ms (%.3g)\n", 
	  tv_toMsecs(&usedSCpu), 
	  (float) tv_toMsecs(&usedSCpu)/tv_toMsecs(&usedTime));

  if (MeasureMe) {
    bwHist = (double*) malloc (((int) (TotalNumRec/Freq)) * sizeof(double));
    if (bwHist == NULL) Styx_errQuit("Not enough resources");
    sensNet_GetHistory(bwHist, (int) (TotalNumRec/Freq), SENSNET_READ);
    fprintf(stdout, "Bandwidth mean:   \t%.3g Mbps (sdev %.3g)\n",
	    Stats_mean (bwHist, (int) (TotalNumRec/Freq)),
	    Stats_sdev (bwHist, (int) (TotalNumRec/Freq)));
    sensNet_Finish();
    free(bwHist);
  }
  fflush(stdout);
} /* end dumpStats */


/*----------------------------------------------------------------------------*/

/*
 * getUsage: Get total CPU usage and wall-clock time point
 *
 * Parameters: Pointer to timeval structs for CPU and Time point
 * 
 * Return: Upon successful completion, 0 is returned.
 *
 */

int
getUsage(struct timeval *pointCpu, struct timeval* pointUCpu, 
	 struct timeval* pointSCpu, struct timeval *pointTime)
{
  struct rusage rusageBuffer;
  
  getrusage(RUSAGE_SELF, &rusageBuffer);
  tv_add2(pointCpu, &(rusageBuffer.ru_utime), &(rusageBuffer.ru_stime));
  tv_assign(pointUCpu, &(rusageBuffer.ru_utime));
  tv_assign(pointSCpu, &(rusageBuffer.ru_stime));
  gettimeofday(pointTime, NULL);
  
  return 0;
} /* end getUsage */


/*----------------------------------------------------------------------------*/

/*
 * interruptHandler: To go out nicely
 *
 * Parameters: signal number
 * 
 * Return: -
 *
 */

static void
interruptHandler(int signo) 
{
  /* If the clock has not been stop after reaching the 
     target number of frames ... */
  if (TotalNumRec == 0)
    getUsage(&stopCpu, &stopUCpu, &stopSCpu, &stopTime);

  dumpStats();  
  EChannel_destroy(chan);
  CManager_close(cm);

  exit(EXIT_SUCCESS);
} /* end interruptHandler */


/*----------------------------------------------------------------------------*/

/*
 * eventHandler:
 *
 * Parameters: 
 * 
 * Return: 
 *
 */

void 
eventHandler (void* event, void* client_data, attr_list attrs)
{
  Raw_data_ptr rawrecP = (Raw_data_ptr) event;
  struct tm *now;
  time_t nowt;
  char snapshotFile[FILENAMELEN];
  char header[PPMHEADERLEN];
  int sfd;

  ++NumRec;
  if (verbose) printf("Got event %d\n", rawrecP->tag);
  time(&nowt);
  now = localtime(&nowt);
  sprintf(snapshotFile, "snapshot%d-%d-%d.ppm", 
	  NumRec, now->tm_hour, now->tm_min);
  sfd = open(snapshotFile, O_CREAT|O_WRONLY, S_IRWXU|S_IRGRP|S_IROTH);
  sprintf(header, "%c%c\n%d %d\n%d\n", rawrecP->ppm1, rawrecP->ppm2, 
	  rawrecP->width, rawrecP->height, 255);  
  write(sfd, header, sizeof(header));
  if (rawrecP->buff != NULL) 
    write(sfd, rawrecP->buff, rawrecP->size);
  close(sfd);
  
  if (rawrecP->tag == TotalNumRec)
    getUsage(&stopCpu, &stopUCpu, &stopSCpu, &stopTime);

  return;
} /* end eventHandler */


/*----------------------------------------------------------------------------*/

int
main(int argc, char* argv[])
{
  EControlContext ec;
  int optc;  
  char* chanID = NULL;

  local_progname = argv[0];
  if (argc < 3) usage(stderr, 1);

  Styx_asDaemon(0);             /* not a daemon, actually do not use syslog */
  Styx_setDebugLevel(0);        /* no debugging unless requested */
  Styx_setProgname(argv[0]);
  while ((optc = getopt_long(argc, argv, "+vhmc:n:s:", 
			     longopts, (int*) 0)) != EOF) {
    switch (optc) {
    case 'v':
      verbose= TRUE;
      Styx_setDebugLevel(7);
      Styx_privateLog(0);
      break;
    case 'h':
      usage(stdout, 0);
      break;
    case 'm':
      MeasureMe = TRUE;
      break;
    case 'c':
      chanID = optarg;
      break;
    case 'n':
      TotalNumRec = atoi(optarg);
      if (TotalNumRec < 0) usage(stderr, 1);
      break;
    case 's':
      Freq = atoi(optarg);
      if (Freq < 0) usage(stderr, 1);
      break;
    default:
      usage(stderr, 1);
    }
  }

  if (optind > argc) usage(stderr, 1); /* no extra args */

  if (verbose) {
    fprintf(stdout, "From channel:      \t%s\n", chanID);
  }

  cm = CManager_create();
  ec = ECho_CM_init(cm);

  if (signal(SIGINT, interruptHandler) == SIG_ERR) 
    Styx_errQuit("Signal error");

  if (verbose) printf("Opening given channel %s\n", chanID);
  chan = EChannel_open(ec, chanID); 
  if (chan == NULL) Styx_errQuit("Failed to open channel\n");

  if (verbose) printf("Subscribing to new channel\n");
  
  if (MeasureMe) {
    sensNet_Init(Freq);
    sensNet_SetHistLen(TotalNumRec/Freq, SENSNET_READ);
  }

  getUsage(&startCpu, &startUCpu, &startSCpu, &startTime);
  ECsink_typed_subscribe(chan, Raw_data_fld, NULL, eventHandler, NULL);

  CMsleep(cm, 6000);		/* 10' and go away  */

  dumpStats();  
  EChannel_destroy(chan);
  CManager_close(cm);

  exit(EXIT_SUCCESS);
} /* main */
