/* -*-C-*-
*******************************************************************************
*
* File:         ebw_source.c
* RCS:          $Id: ebw_source.c,v 1.1.1.1 2003/06/11 18:49:01 wyan Exp $
* Description:  ECho bandwidth experiment
* Author:       Fabian E. Bustamante
*               Systems Research Group
*               College of Computing
*               Georgia Institute of Technology
* Created:      Wed Sep 19, 2001 at 23:01:12
* Modified:     Mon Oct 01, 2001 at 08:24:52 fabianb@cc.gatech.edu
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
* (C) Copyright 2001, Georgia Institute of Technology, all rights reserved.
*
*******************************************************************************
*/

#if HAVE_CONFIG_H
#  include <src/config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <gen_thread.h>
#include <atl.h>
#include <cm.h>
#include <echo.h>
#include <getopt.h>
#include <ebw_common.h>
#include <stats.h>
#include <sensNet.h>

char *local_progname;
static int Measure = 0;

/* Command line arguments */
static struct option longopts[] =
{
  {"verbose", no_argument, NULL, 'v'},
  {"help", no_argument, NULL, 'h'},
  {"measure", no_argument, NULL, 'm'},
  {"iter", required_argument, NULL, 'i'},
  {"lower", required_argument, NULL, 'l'},
  {"upper", required_argument, NULL, 'u'},
  {"step", required_argument, NULL, 's'},
  {"frequency", required_argument, NULL, 'f'},
  {NULL, no_argument, NULL, 0}
};


/*----------------------------------------------------------------------------*/

/*
 * usage: Prints out 'usage' string
 *
 * Parameters: target 'stream' and 'status' to quit with
 *
 * Returns: -
 *
 */

static void 
usage (stream, status)
     FILE* stream; 
     int status;
{  
  fprintf(stream, "Usage: %s [-vh] -i iter -l low -u upper -s step\n"
	  "-v --verbose              \tverbose mode\n"
	  "-h --help                 \tgive this help\n"
	  "-m --measure              \tmeasure this end\n"
	  "-i --iter                 \titerations\n"
	  "-l --lower                \tlower bound\n"
	  "-u --upper                \tupper bound\n"
	  "-s --step                 \tstep per iteration\n"
	  "-f --frequency            \tsampling frequency (> 0)\n"
     	  , local_progname);
  exit(status);
} /* end usage */


/*----------------------------------------------------------------------------*/

int 
main(argc, argv)
     int argc; 
     char* argv[];
{
  CManager cm;
  EChannel chan;
  EControlContext ec;
  ECSourceHandle handle;
  Datarec rec;
  Datarec_ptr recP = &rec;
  int optc, i, j, verbose = 0;
  int xfersize, iteration = 0, lower = 0, upper = 0, step = 0;
  int freq = 0;
  double* bwHist = NULL;
  int totalEvents;

  local_progname = argv[0];
  if (argc < 11) usage(stderr, 1);
  while ((optc = getopt_long(argc, argv, "+vhmi:l:u:s:f:", 
			     longopts, (int*) 0)) != EOF) {
    switch (optc) {
    case 'v':
      verbose = 1;
      break;
    case 'h':
      usage(stdout, 0);
      break;
    case 'm':
      Measure = 1;
      break;
    case 'i':
      iteration = atoi(optarg);
      if (iteration < 0) usage(stderr, 1);
      break;
    case 'l':
      lower = atoi(optarg);
      if (lower < 0) usage(stderr, 1);
      break;
    case 'u':
      upper = atoi(optarg);
      if (upper < 0) usage(stderr, 1);
      break;
    case 's':
      step = atoi(optarg);
      if (step < 0) usage(stderr, 1);
      break;
    case 'f':
      freq = atoi(optarg);
      if (freq < 1) usage(stderr, 1);
      break;
    default:
      usage(stderr, 1);
    }
  }
  
  /* verify arguments */
  if ((lower < 1 || upper < 1) || 
      (lower > upper) || step < 1 || 
      freq < 1) 
    usage(stderr, 1);

  if (optind > argc) usage(stderr, 1); /* no extra args */

  if (verbose) {
    fprintf(stdout, "ECho bandwidth experiment\n"
	    "\tIterations         %d\n"
	    "\tLower bound        %d\n"
	    "\tUpper bound        %d\n"
	    "\tStep               %d\n"
	    "\tSampling frequency %d\n",
	    iteration, lower, upper, step, freq);
  }

  if (upper == lower) totalEvents = iteration;
  else totalEvents = iteration * ((upper - lower) / step);
  if (verbose) fprintf(stdout, "Total events %d\n", totalEvents);  

  gen_pthread_init();
  cm = CManager_create();  
  ec = ECho_CM_init(cm);
  chan = EChannel_typed_create(ec, Datarec_fld, NULL);
  if (chan == NULL) {
    perror("Failed to create channel");
    exit(1);
  }
  
  fprintf(stdout, "Echo channel ID:\n %s\n", ECglobal_id(chan));
  handle = ECsource_typed_subscribe(chan, Datarec_fld, NULL);

  /* Wait for clients */
  CMfork_comm_thread(cm);
  while (EChas_sinks(handle) < 1) 
    CMusleep(cm, 1000);
  if (verbose) fprintf(stdout, "Ready!\n");

  if (Measure) {
    sensNet_Init(freq);
    sensNet_SetHistLen(totalEvents/freq, SENSNET_WRITE);
  }
  
  if (verbose) {
    fprintf(stdout, "Sending\n");
    fflush(stdout);
  }

  xfersize = lower;
  for (i = 0; i < totalEvents; ++i) {
    recP->size = xfersize;
    recP->buff = malloc(xfersize);
    for (j = 0; j < iteration; j++) {
      if (j == (iteration - 1)) recP->tag = 1;
      else recP->tag = 0;
      ECsubmit_typed_event(handle, recP);
    }
    free(recP->buff);
    xfersize += step;
    fprintf(stdout, "Event %d/%d\n", i, totalEvents);
    fflush(stdout);
  }
  if (verbose) fprintf(stdout, "Done sending\n");

  if (Measure) {
    bwHist = (double*) malloc(((int) (totalEvents/freq)) * sizeof(double));
    if (bwHist == NULL) {
      perror("Not enough resources");
      exit(1);
    }
  }

  fflush(stdout);

  if (Measure) {
    sensNet_GetHistory(bwHist, (int) (totalEvents/freq), SENSNET_WRITE);
    for (i = 0; i < (int) (totalEvents/freq); i++) 
      fprintf(stdout, "Bw %.3f Mbps\n", bwHist[i]);

    fprintf(stdout, "Bw mean %.3f Mbps (sdev %.3f)\n", 
	    Stats_mean (bwHist, (int) (totalEvents/freq)),
	    Stats_sdev (bwHist, (int) (totalEvents/freq)));

    fflush(stdout);
    sensNet_Finish();
  
    printf("After finishing\n");
    free(bwHist);
  }

  EChannel_destroy(chan);
  CManager_close(cm);

  exit(0);
} /* end main */

