/* -*-C-*-
*******************************************************************************
*
* File:         styx.c
* RCS:          $Id: styx.c,v 1.1.1.1 2003/06/11 18:49:01 wyan Exp $
* Description:  Error and debug functions
* Author:       Fabian Ernesto Bustamante
*               Systems Research Group
*               College of Computing
*               Georgia Institute of Technology
* Created:      Wed May 30, 2001 at 17:02:28
* Modified:     Mon Jul 09, 2001 at 14:50:49 fabianb@cc.gatech.edu
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
* (C) Copyright 2001, Georgia Institute of Technology, all rights reserved.
*
*******************************************************************************
*/

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>
#include <stdarg.h>		/* ANSI C header file */
#include <syslog.h>		/* for syslog() */
#include <stdlib.h>
#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>

#include <styx.h>

static char *name = NULL;	/* program name for messages */
static int Styx_DaemonProc = 0;	/* is a daemon process ? */
static int Styx_PrivateLog = 1;	/* do we want a private log ? */
static int Styx_debugLevel = 9;
FILE *Styx_Logfp = NULL;
static char *Styx_LogFileName = NULL;

/* 
 * StyxDoit: Print a message and return to caller. 
 *    Caller specifies "errnoflag" and "level". 
 *
 */

void StyxDoit(int errnoflag, int level, const char *fmt, va_list ap);



/*----------------------------------------------------------------------------*/

/*
 * Styx_asDaemon:  Sets/unset the requesting process as a daemon (use syslog 
 *   if it is)
 *
 * Parameters: onOff switch
 *
 * Returns: -
 *
 */

void
Styx_asDaemon(int onOff)
{
  if (onOff) Styx_DaemonProc = 1;
  else {
    Styx_PrivateLog = 0;	/* stdout unless requested otherwise */
    Styx_DaemonProc = 0;
  }
} /* end Styx_asDaemon */


/*----------------------------------------------------------------------------*/

/*
 * Styx_privateLog:  Sets/unset the use of a private log
 *
 * Parameters: onOff switch
 *
 * Returns: -
 *
 */

void
Styx_privateLog(int onOff)
{
  if (Styx_DaemonProc) return;	/* Daemons cannot alter this */

  if (onOff) Styx_PrivateLog = 1;
  Styx_PrivateLog = 0;
} /* end Styx_privateLog */


/*----------------------------------------------------------------------------*/

/*
 * Styx_setLogFile: Sets the name for the log file, the requesting process 
 *   must not be a daemon
 *
 * Parameters: New name for the log file
 *
 * Returns: 0 if successful, -1 otherwise
 *
 */

int  
Styx_setLogFile(char *name)
{
  char *tmpName;

  if (Styx_DaemonProc || (name == NULL)) return -11;

  tmpName = malloc((unsigned)(strlen(name) + 1));
  if (tmpName == NULL) return -1;

  /* Dealloc previous name (better than realloc and clean) */
  if (Styx_LogFileName != NULL) free(tmpName);
  
  /* Assign new path */
  Styx_LogFileName = tmpName;
  strcpy(Styx_LogFileName, name);      
  Styx_PrivateLog = 1;		/* obviously */
  return 0;
} /* end Styx_setLogFile */



/*----------------------------------------------------------------------------*/

/*
 * Styx_setDebugLevel: Sets debug level to the given one 
 *
 * Parameters: Debug level
 *
 * Returns: -
 *
 */

void
Styx_setDebugLevel(int level) 
{
  if (level < 0) Styx_debugLevel = 0;
  else Styx_debugLevel = level;
} /* end Styx_setDebugLevel */


/*----------------------------------------------------------------------------*/

/*
 * Styx_setProgname: Sets stored name of program
 *
 * Parameters: Program name 
 *
 * Returns: -
 *
 */

void 
Styx_setProgname(char *str)
{
  name = (char *) malloc(strlen(str)+1);
  if (name == NULL) {
    StyxDoit(1, LOG_ERR, "%.20s failed - not enough memory.", str);
    exit(1);
  }
  strcpy(name, str);
} /* end Styx_setProgname */


/*----------------------------------------------------------------------------*/

/*
 * Styx_progname: Returns store name of program. 
 *
 * Parameters: -
 *
 * Returns: Stored program name
 *
 */

char *
Styx_progname(void)
{
  return name;
} /* end Styx_progname */


/*----------------------------------------------------------------------------*/

/*
 * Styx_errRet: Handles nonfatal error related to a system call, prints a 
 *   message and return
 *
 * Parameters: Message
 *
 * Returns: -
 *
 */

void
Styx_errRet(const char *fmt, ...)
{
  va_list		ap;

  va_start(ap, fmt);
  StyxDoit(1, LOG_INFO, fmt, ap);
  va_end(ap);
  return;
} /* end Styx_errRet */



/*----------------------------------------------------------------------------*/

/*
 * Styx_errSys: Handles fatal error related to a system call, prints a message 
 *   and terminate
 *
 * Parameters: Message
 *
 * Returns: -
 *
 */

void
Styx_errSys(const char *fmt, ...)
{
  va_list		ap;

  va_start(ap, fmt);
  StyxDoit(1, LOG_ERR, fmt, ap);
  va_end(ap);
  exit(1);
} /* end Styx_errSys */


/*----------------------------------------------------------------------------*/

/*
 * Styx_errDump: Handles fatal error related to a system call, prints a 
 *   message, dump core, and terminate
 *
 * Parameters: Message
 *
 * Returns: -
 *
 */

void
Styx_errDump(const char *fmt, ...)
{
  va_list		ap;

  va_start(ap, fmt);
  StyxDoit(1, LOG_ERR, fmt, ap);
  va_end(ap);
  abort();		/* dump core and terminate */
  exit(1);		/* shouldn't get here */
} /* end Styx_errDump */


/*----------------------------------------------------------------------------*/

/*
 * Styx_errMsg: Handles nonfatal error unrelated to a system call, prints a 
 *   message and return 
 *
 * Parameters: Message
 *
 * Returns: -
 *
 */

void
Styx_errMsg(const char *fmt, ...)
{
  va_list		ap;

  va_start(ap, fmt);
  StyxDoit(0, LOG_INFO, fmt, ap);
  va_end(ap);
  return;
} /* end Styx_errMsg */


/*----------------------------------------------------------------------------*/

/*
 * Styx_errQuit: Handles fatal error unrelated to a system call, prints a 
 *   message and terminate
 *
 * Parameters: Message
 *
 * Returns: -
 *
 */

void
Styx_errQuit(const char *fmt, ...)
{
  va_list		ap;

  va_start(ap, fmt);
  StyxDoit(0, LOG_ERR, fmt, ap);
  va_end(ap);
  exit(1);
} /* end Styx_errQuit */


/*----------------------------------------------------------------------------*/

/*
 * Styx_debugMsg: Prints debug-level message and return to caller
 *
 * Parameters: Level of the message and message
 *
 * Returns: -
 *
 */

void
Styx_debugMsg(int debugLevel, const char *fmt, ...)
{
  va_list		ap;

  if (debugLevel < Styx_debugLevel) {
    va_start(ap, fmt);
    StyxDoit(0, LOG_DEBUG + debugLevel, fmt, ap);
    va_end(ap);
  }
  return;
} /* end Styx_debugMsg */


/*----------------------------------------------------------------------------*/

/*
 * StyxDoit: Prints a message and return to caller.
 *
 * Parameters: "errnoflag" and "level" plust the message and its format
 *
 * Returns: -
 *
 */

void
StyxDoit(int errnoflag, int level, const char *fmt, va_list ap)
{
  int errno_save, n;
  char buf[MAXLINE];
  char tmpName[MAXFILENAME];
  
  errno_save = errnoflag; /* value caller might want printed */

  /* vsnprintf(buf, sizeof(buf), fmt, ap);     this is safe  */

  vsprintf(buf, fmt, ap);
  
  n = strlen(buf);
  if (errnoflag) {
    /* snprintf(buf+n, sizeof(buf)-n, ": %s",  */
    /*     strerror(errno_save)); this is safe */
    sprintf(buf+n, ": %s", strerror(errno_save));
  }
  
  strcat(buf, "\n");

  if (Styx_DaemonProc) {
    syslog(level, buf);
  } else {
    if (Styx_PrivateLog) {
      if(Styx_Logfp == NULL) {
	if (!Styx_LogFileName) {
	  sprintf(tmpName, "/tmp/LogFile-%u.log", (u_int) getpid());
	  Styx_setLogFile(tmpName);
	}
	Styx_Logfp = fopen(Styx_LogFileName, "w");
	if (Styx_Logfp == NULL) {
	  Styx_PrivateLog = 0;
	  return;
	}
      }
      fputs(buf, Styx_Logfp);
    } else {
      fflush(stdout);		/* in case stdout and stderr are the same */
      fputs(buf, stderr);
      fflush(stderr);
    }
  }
  
  return;
} /* end StyxDoit */

/*----------------------------------------------------------------------------*/
