/* -*-C-*-
*******************************************************************************
*
* File:         tv.c
* RCS:          $Id: tv.c,v 1.1.1.1 2003/06/11 18:49:01 wyan Exp $
* Description:  Timeval associated functions
* Author:       Fabian Ernesto Bustamante
*               Systems Research Group
*               College of Computing
*               Georgia Institute of Technology
* Created:      Wed Feb 28, 2001 at 17:28:43
* Modified:     Mon Jul 09, 2001 at 15:07:17 fabianb@cc.gatech.edu
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
* (C) Copyright 2001, Georgia Institute of Technology, all rights reserved.
*
*******************************************************************************
*/

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <sys/time.h>
#include <tv.h>

/*----------------------------------------------------------------------------*/

/*
 * tv_zero: Sets struct timeval var
 *
 * Parameters: Pointer to struct timeval var
 *
 * Returns: 0 if success, -1 otherwise
 *
 */

int 
tv_zero(t0)
     struct timeval* t0; 
{
  if (t0 == NULL) return -1;
  t0->tv_sec = 0; 
  t0->tv_usec = 0;
  return 0;
} /* end tv_zero */


/*----------------------------------------------------------------------------*/

/*
 * tv_assign: Assigns 2nd struct timeval var to first one
 *
 * Parameters: Pointers to struct timeval vars
 *
 * Returns: 0 is success, -1 otherwise
 *
 */

int
tv_assign(t0, t1)
     struct timeval* t0;
     struct timeval* t1;
{
  if ((t0 == NULL) || (t1 == NULL)) return -1;
  t0->tv_sec = t1->tv_sec;
  t0->tv_usec = t1->tv_usec;
  return 0;
} /* end tv_assign */


/*----------------------------------------------------------------------------*/

/*
 * tv_toSecs: Translates to secs
 *
 * Parameters: Pointer to struct timeval var
 *
 * Returns: Timeval in secs, -1 otherwise
 *
 */

u_long
tv_toSecs(t0)
     struct timeval* t0;
{  
  if (t0 == NULL) return -1;
  return (t0->tv_sec + t0->tv_usec*1e-6);
} /* end tv_toSecs */


/*----------------------------------------------------------------------------*/

/*
 * tv_toMsecs: Translates to msecs
 *
 * Parameters: Pointer to struct timeval var
 *
 * Returns: Timeval in msecs, -1 otherwise
 *
 */

u_long
tv_toMsecs(t0)
     struct timeval* t0;
{    
  if (t0 == NULL) return -1;
  return (t0->tv_sec*1e3 + t0->tv_usec*1e-3);
} /* end tv_toMsecs */


/*----------------------------------------------------------------------------*/

/*
 * tv_toUsecs: Translates to microsecs
 *
 * Parameters: Pointer to struct timeval var
 *
 * Returns: Timeval in msecs, -1 if error
 *
 */

u_long
tv_toUsecs(t0)
     struct timeval* t0;
{ 
  if (t0 == NULL) return -1;
  return (t0->tv_sec*1e6 + t0->tv_usec);
} /* end tv_toMsecs */


/*----------------------------------------------------------------------------*/

/*
 * tv_msecsTo: Translates msecs to struct timeval var
 *
 * Parameters: Time in msecs and pointer to struct timeval var 
 *
 * Returns: 0 if success, -1 otherwise
 *
 */

int 
tv_msecsTo(ms, out)
     u_long ms; 
     struct timeval* out;
{
  if (out == NULL) return -1;
  out->tv_sec = ms*1e-3;
  out->tv_usec = (ms - (out->tv_sec * 1e3))*1e3;
  return 0;
} /* end tv_msecsTo */


/*----------------------------------------------------------------------------*/

/*
 * tv_minus2: Substracts t2 from t1 and returns results in t0
 *
 * Parameters: Pointers to result and operators struct timeval vars
 *
 * Returns: 0 if success, -1 otherwise
 *
 */
int
tv_minus2(t0, t1, t2)
     struct timeval* t0; 
     struct timeval* t1;
     struct timeval* t2;
{  
  if ((t0 == NULL) || (t1 == NULL) || (t2 == NULL)) return -1;
  t0->tv_sec = t1->tv_sec - t2->tv_sec;
  t0->tv_usec = t1->tv_usec - t2->tv_usec; 

  while (t0->tv_usec < 0) {
    t0->tv_sec--;
    t0->tv_usec += 1e6;
  }
  return 0;
} /* end tv_minus2 */


/*----------------------------------------------------------------------------*/

/*
 * tv_minus:  Substracts t1 from t0 and returns results in t0
 *
 * Parameters: Pointers to result and operators struct timeval vars
 *
 * Returns: 0 if success, -1 otherwise
 *
 */

int
tv_minus(t0, t1)
     struct timeval* t0; 
     struct timeval* t1;
{
  if ((t0 == NULL) || (t1 == NULL)) return -1;
  t0->tv_sec -= t1->tv_sec;
  t0->tv_usec -= t1->tv_usec;
  
  while (t0->tv_usec < 0) {
    t0->tv_sec--;
    t0->tv_usec += 1e6;
  }
  return 0;
} /* end tv_minus */


/*----------------------------------------------------------------------------*/

/*
 * tv_add2: Add t1 and t2, returns results in t0
 *
 * Parameters: Pointers to result and operators struct timeval vars
 *
 * Returns: 0 if success, -1 otherwise
 *
 */

int
tv_add2(t0, t1, t2)
     struct timeval* t0; 
     struct timeval* t1; 
     struct timeval* t2;
{
  if ((t0 == NULL) || (t1 == NULL) || (t2 == NULL)) return -1;
  t0->tv_sec = t1->tv_sec + t2->tv_sec;
  t0->tv_usec = t1->tv_usec + t2->tv_usec;
  
  while (t0->tv_usec >= 1e6) {
    t0->tv_sec++;
    t0->tv_usec -= 1e6;
  }  
  return 0;
} /* end tv_add2 */


/*----------------------------------------------------------------------------*/

/*
 * tv_add: Add t0 and t1, returns result in t0
 *
 * Parameters: Pointers to result and operators struct timeval vars
 *
 * Returns: 0 if success, -1 otherwise
 *
 */

int
tv_add(t0, t1)
     struct timeval* t0; 
     struct timeval* t1;
{  
  if ((t0 == NULL) || (t1 == NULL)) return -1;
  t0->tv_sec += t1->tv_sec;
  t0->tv_usec += t1->tv_usec;
  
  while (t0->tv_usec >= 1e6) {
    t0->tv_sec++;
    t0->tv_usec -= 1e6;
  }  
  return 0;
} /* end tv_add */
