// out.cpp - implementation of functions to output the DOM document 
//
#include <ostream.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <stdarg.h>

#include "isg.hpp"
#include "out.hpp"
#include "outAllNodes.hpp"

// some information that should be available everywhere
struct isgOutStruct {
  // the template we are going to fill each time
  DOM_Element templt;

  // our maps of types from a pipe type (e.g. "integer")
  // to a language type (e.g. for C "int")
  DOM_NodeList typeMaps;

} isgOut;



//////////////////////////////////////////////////////////////////////////////////////////
// takes a template and nodeSet and fills in the template
int generateCode(DOM_Document templateDocument, DOM_Document nodeSetDocument)
{
  // get template pertaining to allNodes from teh template XML doc
  // get list of all allNodes sections
  DOM_NodeList nodeList;
  DOM_Element allNodesTemplate;
  DOM_Element perNodeTemplate;

  
  isgOut.templt = templateDocument.getDocumentElement();

  // get the list of type maps we need
  isgOut.typeMaps = templateDocument.getDocumentElement().getElementsByTagName("mapTable");

  // generate code for allNodes
  // assuming there is only one <allNodes> in the template file
  // if there is more than two <allNodes>, use the first one
  nodeList = templateDocument.getElementsByTagName("allNodes");
  if( nodeList.getLength() >= 1 )
  {
    // use first one
    allNodesTemplate = DOM_ELEMENT(nodeList.item(0));

    generateAllNodes( allNodesTemplate, nodeSetDocument );
  }

  return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
// prototypes for ref of local funcs

// for printing to a file
static int textprint2stream( FILE *f, DOM_Element r );

// some utility functions
char *getTextOfNode(char *, DOM_Node node);
int makePathDir(char *path, mode_t mode);
DOMString mapType( DOMString fromType, DOMString toDomain );

int fillTagWithText(DOM_Element root, char *tagName, DOMString text)
{
  DOM_NodeList nodeList;

  nodeList = root.getElementsByTagName(tagName);
  for(int i=0; i < nodeList.getLength(); i++)
  {
    //makeText(nodeList.item(i), text);
    //Temporary : young   
    makeTextAndMarkNode(nodeList.item(i), text);
  }
  return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
// mapType() -- maps the XML schema type to a PBIO type or C type
//           -- we do both in one function to facilitate ensuring that
//           -- the mapping is 1:1 between PBIO and C types
//           -- returns a DOMString with the desired info
//           -- or NULL on failure
DOMString mapType( DOMString fromType, DOMString toDomain )
{
  DOM_NodeList nl;
  DOM_Node tnode;
  DOM_Element mapNode;
  int i = 0, max;
 
//  cout << "out.cpp::mapType():enter   ";
//  fromType.print(); 
//  cout << "      ";
//  toDomain.print();
//  cout << endl;
  max = isgOut.typeMaps.getLength();
  for( i = 0; i < max; i++ )
    {
    // convert to something we can get attribute from
    tnode = isgOut.typeMaps.item(i);
    mapNode = (DOM_Element &) tnode;
    // see if this is the right mapping structure
//    cout << " trying:    ";
//    mapNode.getAttribute("to").println();
    if( mapNode.getAttribute("to").equals( toDomain ) ) 
      {
      break;
      }
    }
  // fail to find match? return failure!
  if( i >= max )
    return NULL;

  // otherwise, continue on and look up the proper mapping
  // by getting  list of the mapping lines and checking ofr
  // another match
  nl = mapNode.getElementsByTagName("mapping");

  max = nl.getLength();
  for( i = 0; i < max; i ++)
    {
    tnode = nl.item(i);
    mapNode = (DOM_Element &) tnode;
    // check for it
    if( mapNode.getAttribute("from").equals( fromType ) ) 
      {
      // found it!!! return it
      return mapNode.getAttribute("to");
      }
    }

  // must not have found it, return failure code
  return NULL;
}

int makeText( DOM_Node futureParent, DOMString textToAdd )
{
  DOM_Element fParent;
  DOM_Text ttext;

//  cout << "out.cpp: makeText(): enter" << endl;
  fParent = (DOM_Element &) futureParent;
  // create and append the text node
  ttext = futureParent.getOwnerDocument().createTextNode( textToAdd );
  futureParent.appendChild( ttext );

//  cout << "out.cpp: makeText(): exit" << endl;
  return 0;
}

// given a parent node in the tree, we need to mark it for printing
// this way we can print "clean" code without having all the other
// cruft in the way
int makeTextAndMarkNode( DOM_Node futureParent, DOMString textToAdd )
{
  DOM_Element fParent;
  DOM_Text ttext;

//  cout << "out.cpp: makeTextAndMarkNode(): enter" << endl;
  fParent = (DOM_Element &) futureParent;
  // create and append the text node
  ttext = futureParent.getOwnerDocument().createTextNode( textToAdd );
  futureParent.appendChild( ttext );

  // mark the parent
  if( fParent.getAttribute("mark").equals("") )
    {
    // if un-marked, we need to mark it
    markElement( fParent );
    }
  
//  cout << "out.cpp: makeTextAndMarkNode(): exit" << endl;
  return 0;
}

int isElementMarked( DOM_Element n )
{
  return n.getAttribute("mark").equals("true");
}
int markElement( DOM_Element n )
{
  //if( n.getNodeName().equals("aspect") ) cout << "markElement():marking aspect" << endl;
  n.setAttribute("mark","true");
}

// updates the print markings on all nodes, a node is "marked" if
// it has any node beneath it that is marked
// when a node is marked, all the text nodes beneath it will be marked
int updateMarks( DOM_Element e )
{
  DOM_Node tnode;
  DOM_Text texnode;
  DOM_Element e2;
  int thisMarked = 0;

  // if we're already marked, we must still update sub-tree, but
  // we don't have to check each node as it completes
  if( isElementMarked(e))
    thisMarked = 1;

  // check children first
  tnode = e.getFirstChild();
  while( tnode != NULL )
    {
      if( tnode.getNodeType() == DOM_Node::ELEMENT_NODE )
        {
        e2 = (DOM_Element &)tnode;
        // update the child
        updateMarks( e2 );
        
        // if we end up with a marked child, then we must 
        // mark this node if not already done. Keep track of whether 
        // we already have decided to mark it for speed.
        if( !thisMarked && isElementMarked(e2) )
          {
          // set the marking
          thisMarked = 1;
          markElement(e);
          }
        }
      tnode = tnode.getNextSibling();
    }
}

