//////////////////////////////////////////////////////////////////////////////////////////
// outAllNodes.cpp - implementation of functions to output the DOM document 
//
#include <fstream.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "isg.hpp"
#include "out.hpp"
#include "outPerPipe.hpp"

// these functions fill out the parts of a template having to do with
// data definitions
static int generateDataDefCode(DOM_Element templt, DOM_Document nodeSetDocument);
static int fillDataDefCode(DOM_Element templt, DOM_NodeList dataDefDescList) ;
static int fillDataTemplate( DOM_Element dataTemplate, DOM_Element dataSource );
static int fillDataArgTemplate( DOM_Element templ, DOM_Element argData );
static int fillDataArrayArgTemplate( DOM_Element templ, DOM_Element argData );

// this will fill the part of a template that has to do with filter
// definitions
static int generateFilterCode(DOM_Element templt, DOM_Document nodeSetDocument);
static int fillFilterCode(DOM_Element templt, DOM_Element nodeSetDocument);

/*
  This functions fills up the template with meaningful data taken from
  a description file.
  Return:        0 Successful
                 Non-zero Error occurred (May also print error to stderr)
*/
int generateAllNodes( DOM_Element allNodesTemplate, DOM_Document nodeSetDocument )
{
  DOM_NodeList nodeList;
  DOM_Element dataDefDesc;
  int dataDefTemplateNum;
  DOM_Element *dataDefTemplate;
  
  // process <dataDef>
  generateDataDefCode(allNodesTemplate, nodeSetDocument);

  // process <perPipe>
  generatePerPipeCode(allNodesTemplate, nodeSetDocument);

  // update to print new stuff
  updateMarks( allNodesTemplate.getOwnerDocument().getDocumentElement() );

  
  return 0;
}

static int generateDataDefCode(DOM_Element templt, DOM_Document nodeSetDocument)
{
  DOM_Element dataDefDesc;
  int dataDefTemplateNum;
  DOM_Element *dataDefTemplate;
  DOM_NodeList nDataDefList;
  DOM_NodeList tDataDefList;

  // process <dataDef>
  
  // we will find <dataDef> from dataDefDesc
  dataDefDesc = DOM_ELEMENT(nodeSetDocument.getDocumentElement());

  // get all of our dataDef descriptions
  nDataDefList = dataDefDesc.getElementsByTagName("dataDef");

  // publish pipe description to file
  if (gFlags & FLAG_PUBLISH_FILE)
  {
    char path[128];
    char *publishRoot;

    // find publishing root and if not, use pwd
    if ((publishRoot = getenv("INFOPIPE_PUBLISH_ROOT")) == NULL)
    {
      publishRoot = getenv("PWD");
    }

    // set up the path to the repository file
    makePathString(path, publishRoot, "dataDef", NULL);
    // check and make path if necessary
    makePathDir(path);

    for (int i=0; i < nDataDefList.getLength(); i++)
    {
      char *dataDefName = DOM_ELEMENT(nDataDefList.item(i)).getAttribute("dataTypeName").transcode();
      char filename[128];

      // file name will be dataDef/dataDefDescription
      sprintf(filename, "%s%sDescription", path, dataDefName);

      ofstream ofile(filename);
      ofile << DOM_ELEMENT(nDataDefList.item(i));

      delete dataDefName;
    }
  }    

  // get all of <dataDef> in the template
  tDataDefList = templt.getElementsByTagName("dataDef");
  dataDefTemplateNum = tDataDefList.getLength();
  
  // because tDataDefList change as dataDef node clones and appends
  // we copy original nodes to some array
  dataDefTemplate = new DOM_Element[dataDefTemplateNum];
  // and fill entire array (need casting)
  for(int i=0; i < dataDefTemplateNum; i++)
  {
    dataDefTemplate[i] = DOM_ELEMENT(tDataDefList.item(i));
  }

  // for all of <dataDef> in the template
  for(int i=0; i < dataDefTemplateNum; i++)
  {
    // fill this out
    fillDataDefCode(dataDefTemplate[i], nDataDefList);
  }

  delete [] dataDefTemplate;

  return 0;
}

// fills in information relating to data types in the main 
// document irrespective of whether the data type info needs
// to be filled in in the nodes or for the entire pipe.
static int fillDataDefCode(DOM_Element templt, DOM_NodeList dataDefDescList) 
{
  DOM_Element newTemplate, currentDataDef ;
  int i,len;


  // for every template, we must create enough copies to
  // for all of our data definitions
  for( i = 0; i < dataDefDescList.getLength(); i++)
  {
    // clone the template node and its sub-nodes (deep copy)
    newTemplate = DOM_ELEMENT(templt.cloneNode( true ));

    // attach to the same parent as original template
    // insert right before the original template
    templt.getParentNode().insertBefore( newTemplate, templt );

    // fill in the info
    currentDataDef = DOM_ELEMENT(dataDefDescList.item(i));
    fillDataTemplate( newTemplate, currentDataDef );
  }

  // remove our original template
  // and then we can move to the next template
  templt.getParentNode().removeChild(templt);

  return 0;
}

static int fillDataTemplate( DOM_Element dataTemplate, DOM_Element dataSource )
{
  DOM_Element *dataArgTemplate;
  DOM_NodeList nodeList;
  int dataArgTemplateNum;
  DOM_Attr newAttr;

  int i;

  newAttr =  dataTemplate.getOwnerDocument().createAttribute("num");
  newAttr.setValue( dataSource.getAttribute("num") );
  dataTemplate.setAttributeNode( newAttr );


  // process <name> tag
  // since all <name> element will be processed, 
  // no further functions will worry about <name>
  DOM_NodeList nameList = dataTemplate.getElementsByTagName("dataTypeName");
  for (i = 0; i < nameList.getLength(); i++)
  {
    makeTextAndMarkNode(DOM_ELEMENT(nameList.item(i)), dataSource.getAttribute("name"));
  }

  // (young) I changed this routine because <dataArg> can be below <aspect>.
  // at that case, the original code seems not to work..

  //////////////////////////////////////////////
  // process <dataArg>

  // get all of <dataArg> in the template
  nodeList = dataTemplate.getElementsByTagName("dataArg");
  dataArgTemplateNum = nodeList.getLength();
  
  // because nodeList change as dataArg node clones and appends
  // we copy original nodes to some array
  dataArgTemplate = new DOM_Element[dataArgTemplateNum];
  for(i = 0; i < dataArgTemplateNum; i++)
  {
    dataArgTemplate[i] = DOM_ELEMENT(nodeList.item(i));
  }

  // for all of <dataArg> in the template
  for(i = 0; i < dataArgTemplateNum; i++)
  {
    DOM_NodeList argList;
    DOM_Element	newTemplate, currTemplate;

    argList = dataSource.getElementsByTagName("arg");
    currTemplate = dataArgTemplate[i];

    for(int j = 0; j < argList.getLength(); j++ )
    {
      // clone the template
      newTemplate = DOM_ELEMENT(currTemplate.cloneNode( true ));

      // add it to the same parent
      currTemplate.getParentNode().insertBefore( newTemplate, currTemplate );

      // process the new template
      fillDataArgTemplate( newTemplate, DOM_ELEMENT(argList.item(j)) );
    }

    // remove original template
    currTemplate.getParentNode().removeChild(currTemplate);
  }

  delete [] dataArgTemplate;


  //////////////////////////////////////////////
  // process <dataArrayArg>

  DOM_Element *dataArrayArgTemplate;
  int dataArrayArgTemplateNum;

  // get all of <dataArrayArg> in the template
  nodeList = dataTemplate.getElementsByTagName("dataArrayArg");
  dataArrayArgTemplateNum = nodeList.getLength();
  
  // because nodeList change as dataArrayArg node clones and appends
  // we copy original nodes to some array
  dataArrayArgTemplate = new DOM_Element[dataArrayArgTemplateNum];
  for(i = 0; i < dataArrayArgTemplateNum; i++)
  {
    dataArrayArgTemplate[i] = DOM_ELEMENT(nodeList.item(i));
  }

  // for all of <dataArrayArg> in the template
  for(i = 0; i < dataArrayArgTemplateNum; i++)
  {
    DOM_NodeList arrayArgList;
    DOM_Element	newTemplate, currTemplate;

    arrayArgList = dataSource.getElementsByTagName("arrayArg");
    currTemplate = dataArrayArgTemplate[i];

    for(int j = 0; j < arrayArgList.getLength(); j++ )
    {
      // clone the template
      newTemplate = DOM_ELEMENT(currTemplate.cloneNode( true ));

      // add it to the same parent
      currTemplate.getParentNode().insertBefore( newTemplate, currTemplate );

      // process the new template
      fillDataArrayArgTemplate( newTemplate, DOM_ELEMENT(arrayArgList.item(j)) );
    }

    // remove original template
    currTemplate.getParentNode().removeChild(currTemplate);
  }

  delete [] dataArrayArgTemplate;

  return 0;
}

// fills in a dataArg line in a dataDef section of a template
static int fillDataArrayArgTemplate( DOM_Element templ, DOM_Element argData )
{
  DOM_Node tNode;
  DOM_Text textNode;
  DOM_Element current,parent;

//  cout << "out.cpp::fillDataArrayArgTemplate():enter" << endl;
  if( argData == NULL )
  {
    cerr << "out.cpp::fillDataArrayArgTemplate(): argData argument is NULL." << endl;
    return 1;
  }
  if( templ == NULL )
    {
    cerr << "Template null." << endl;
    return 1;
    }
  tNode = templ.getFirstChild();
  current = (DOM_Element &)tNode;
  while( current !=NULL )
    {
    if(current.getNodeName().equals("argName") )
      {
      // the name of this argument
//      cout << "fillDataArrayArgTemplate():argName: ";
//      argData.getAttribute("name").print();
//      cout << "...making text node with name.";
      makeTextAndMarkNode( current, argData.getAttribute("name") );
//      cout << "...done." << endl;
//      textNode = current.getOwnerDocument().createTextNode(
//                                              argData.getAttribute("name") );
//      current.appendChild(textNode);
      }
    else if(current.getNodeName().equals("arraySize") )
      {
      // the name of this argument
//      cout << "fillDataArrayArgTemplate():size: ";
//      argData.getAttribute("size").print();
//      cout << "...making text node with size.";
      makeTextAndMarkNode( current, argData.getAttribute("size") );
//      cout << "...done." << endl;
//      textNode = current.getOwnerDocument().createTextNode(
//                                              argData.getAttribute("size") );
//      current.appendChild(textNode);
      }
    else if(current.getNodeName().equals("mapType") )
      {
      // we have to do type conversion on these and map the Infopipe
      // type to a language specific type
//      cout << "fillDataArrayArgTemplate():mapType:"; 
//      argData.getAttribute("type").print();
//      cout << "     to:";
//      current.getAttribute("to").print();
//      mapType(argData.getAttribute("type"), current.getAttribute("to")).println();
      
      makeTextAndMarkNode( current, 
                             mapType(argData.getAttribute("type"), 
                                     current.getAttribute("to")    ) );
//      textNode = current.getOwnerDocument().createTextNode(
//                            mapType(argData.getAttribute("type"), 
//                                  current.getAttribute("to") )  );
//      current.appendChild(textNode);
      }
    else
      {
      }
    // go to the next tag to be filled in
    tNode = current.getNextSibling();
    current = (DOM_Element &)tNode;
    }
//  cout << "out.cpp::fillDataArrayArgTemplate():leave" << endl;
}

// fills in a datArg line in a dataDef section of a template
static int fillDataArgTemplate( DOM_Element templ, DOM_Element argData )
{
  DOM_Node tNode;
  DOM_Text textNode;
  DOM_Element current,parent;

  if( argData == NULL )
  {
    cerr << "out.cpp::fillDataArgTemplate(): argData argument is NULL." << endl;
    return 1;
  }
//  cout << "out.cpp::fillDataArgTemplate():enter" << endl;
  if( templ == NULL )
    {
    cerr << "Template null." << endl;
    return 1;
    }
  tNode = templ.getFirstChild();
  current = (DOM_Element &)tNode;
  while( current !=NULL )
    {
    if(current.getNodeName().equals("argName") )
      {
      // the name of this argument
//      cout << "fillDataArgTemplate():argName: ";
//      argData.getAttribute("name").print();
//      cout << "...making text node with name.";
      makeTextAndMarkNode( current, argData.getAttribute("name") );
//      cout << "...done." << endl;
//      textNode = current.getOwnerDocument().createTextNode(
//                                              argData.getAttribute("name") );
//      current.appendChild(textNode);
      }
    else if(current.getNodeName().equals("mapType") )
      {
      // we have to do type conversion on these and map the Infopipe
      // type to a language specific type
//      cout << "fillDataArgTemplate():mapType:"; 
//      argData.getAttribute("type").print();
//      cout << "     to:";
//      current.getAttribute("to").print();
//      mapType(argData.getAttribute("type"), current.getAttribute("to")).println();
      
      makeTextAndMarkNode( current, 
                             mapType(argData.getAttribute("type"), 
                                     current.getAttribute("to")    ) );
//      textNode = current.getOwnerDocument().createTextNode(
//                            mapType(argData.getAttribute("type"), 
//                                  current.getAttribute("to") )  );
//      current.appendChild(textNode);
      }
    else
      {
      }
    // go to the next tag to be filled in
    tNode = current.getNextSibling();
    current = (DOM_Element &)tNode;
    }
//  cout << "out.cpp::fillDataArgTemplate():leave" << endl;
}


