/* 
 * PPMP.h: the fragment class of PPM frame
 */

#ifndef _PPMP_H
#define _PPMP_H

extern "C" {
#include <string.h>
#include <netinet/in.h>
}

#include <iostream>


typedef unsigned long long uint64;
typedef   signed long long  int64;
typedef unsigned long      uint32;
typedef   signed long       int32;
typedef unsigned short     uint16;
typedef   signed short      int16;
typedef unsigned char      uint8;
typedef   signed char       int8;
 
inline uint16 ChangeEndian16(uint16 v) { return (v>>8) | (v<<8); }
inline uint32 ChangeEndian32(uint32 v) { return (v>>24) | ((v>>8)&0xFF00) | ((v&0xFF00)<<8) | (v<<
24); }

#ifdef WORDS_BIGENDIAN
inline uint16 ToLittleEndian16(uint16 v) { return ChangeEndian16(v); }
inline uint32 ToLittleEndian32(uint32 v) { return ChangeEndian32(v); }
inline uint16 ToBigEndian16(uint16 v)    { return v; }
inline uint32 ToBigEndian32(uint32 v)    { return v; }
#else
inline uint16 ToLittleEndian16(uint16 v) { return v; }
inline uint32 ToLittleEndian32(uint32 v) { return v; }
inline uint16 ToBigEndian16(uint16 v)    { return ChangeEndian16(v); }
inline uint32 ToBigEndian32(uint32 v)    { return ChangeEndian32(v); }
#endif


//The characteristics of each PPM frame.
//This infomation is needed both at the PPM frame
//Sender and Receiver side.
typedef struct {
  int tag;          //Reserved
  char ppm1;        //'P'
  char ppm2;        //'1'-'6'
  int size;         //width*height*pixel_size_in_bytes
  int width;
  int height;
  char* buff;       //Reserved
} PPM_frame_info, *PPM_frame_info_ptr;


  /* Send out the whole PPM frame in the following order/frag:
     PPMP_FRAME_HEADER 4bytes + PPMP_FRAME_NO 4bytes + PPM_frame_info 24bytes
     PPMP_PICT_HEADER  4bytes + PPMP_FRAME_NO 4bytes + PPM_Pixel_Data ? bytes
     PPMP_PICT_HEADER  4bytes + PPMP_FRAME_NO 4bytes + PPM_Pixel_Data ? bytes
     ...
     REPEAT several times to send the whole frame.
  */

//#define MUS 15360
//Only used in sending PPM_frame_info before sending
//each fragment of the whole PPM frame
#define PPMP_FRAME_HEADER_OFFSET 0
#define PPMP_FRAME_INFO_OFFSET 8
//Used in sending each fragment of the whole PPM frame
#define PPMP_PICT_HEADER_OFFSET 0
#define PPMP_FRAME_NO_OFFSET 4

//Used as to distinguished from Mpeg headers in DVDview
#define PPMP_FRAME_HEADER 0x1ff
#define PPMP_PICT_HEADER 0x1f0

class PPMP_Fragment 
{
public:
  PPMP_Fragment(char *frag_buf, size_t frag_len);

  //ppmp_header is either the frame header or the picture header;
  //msg_buf includes either the PPM_frame_info or fragment of PPM
  //pixel data => len varies.
  //
  //len: the msg length without header
  //frag_buf: the pointer to the fragment passed from the caller.
  PPMP_Fragment(unsigned long ppmp_header,
                unsigned long frame_no ,
                const char *msg_buf, size_t len,
                char *frag_buf);

  size_t get_frag_length()
    {
      return frag_length;
    }

  size_t get_payload_length()
    {//No matter it's a frame or pict header, data starts from 
      //PPMP_FRAME_INFO_OFFSET
      return frag_length - PPMP_FRAME_INFO_OFFSET;
    }

  void copy_payload(char *target_buffer)
    {
      memcpy(target_buffer, frag_buffer + PPMP_FRAME_INFO_OFFSET, 
	     frag_length - PPMP_FRAME_INFO_OFFSET);
    }

  unsigned long get_ppmp_frame_header()
    {
      return ntohl(*((unsigned long *) (frag_buffer + PPMP_FRAME_HEADER_OFFSET)));
    }

  unsigned long get_ppmp_pict_header()
    {
      return ntohl(*((unsigned long *) (frag_buffer + PPMP_PICT_HEADER_OFFSET)));
    }

  unsigned long get_ppmp_frame_no()
    {
      return ntohl(*((unsigned long*) (frag_buffer+ PPMP_FRAME_NO_OFFSET)));
    }

protected:
  //All the frame frag info including the necessary header
  char *frag_buffer;
  //sizeof(frag_buffer)
  size_t frag_length;
};


//////////////////////////////////////////////////////////////////
  
class PPMP_Sender {
 public:
  //PPMP_Sender();
  //mus: the pre-defined maximum packet size
  //pixelsize: in bytes, 2 or 3 or 4 perpixel, here's 3
  //maxpixelperline: # of pixels each line of the widest PPM frame
  PPMP_Sender(int mus, int pixelsize, int maxpixelperline
	      /*, 
		PPM_frame_info& ppmfi*/);

  //Send the PPM_frame_info first
  int send_preframe(char* PPMHeader);

  //Then the fragments of the whole frame.
  //PPMFrame_num: the current PPM frame number
  int send_frame(char *msg_buf, size_t len, unsigned long PPMFrame_num);

  virtual int send_fragment(const char *frag_buf, size_t len)=0;
  
  virtual void setPPMFrameInfo(int mus, PPM_frame_info& ppmfi) = 0;

  // private:
  PPM_frame_info pfi;
  int bpp;          //color depth
  int max_buf_size; //the negotiated largest buffer size
  int lines_send_per_time;//# of lines sent each fragment
  int num_times;//# of fragments sent

 private:
  char* fragment_buffer;//passed from the caller
  unsigned long curr_frame_num;
};


//////////////////////////////////////////////////////////////////

//This class used in PPMP_Receiver for accumulating info

class Flag_Array {
 public:
  Flag_Array();
  Flag_Array(int first, int last);
  //Flag_Array(const Flag_Array&);

  int get_current_idx() {return current;}
  int get_start_idx() {return start;}
  int get_end_idx() {return end;}
  void increase_idx(int i) {current+=i;}
  void set_available(int idx) 
    {
    available[idx] = true;}

  void init_available();

  bool all_available() {
    bool b = false;
    for (int i=start; i<=end; i++) {
      if (available[i]==false) 

	return b;
    }
    return true;
  }

 private:
  //line # 0 in the PPM frame
  int start;
  //line # largest in the PPM frame
  int end;
  //pointing to the current accumulated line #
  int current;
  //array of flags indicating if the specific line is accumulated
  bool* available;
};

///////////////////////////////////////////////////////////////

class PPMP_Receiver {
 public:
  //mus : used in negotiating frag size
  PPMP_Receiver(int mus);
  //~PPMP_Receiver();

  //msg: the PPM frame fragment received, including header  
  int process_ppm(const char* msg, size_t len, 
			  unsigned long PPMFrameNum,
			  PPM_frame_info& pfi);

  //Get the necessary info of the receiving frame first.
  //Img: The PPM_FRAME_HEADER info which includes PPM_frame_info.
  //first: line # 0 in a frame
  //last: line # maximum in a frame
  //d_start_line: line # 0 in a frag
  //d_end_line: line # maxmimum in a frag
  //i_start_line: current frag put into buffer starting from this line
  void start_defragment(const char* img, int first,int last,
				int d_start_line, int d_end_line,
				int i_start_line);

  //frag_buf: 
  //frag_len:
  //flushOrNot: piece of accumulated img needs to be discarded if 
  //            bad thing happens.
  char* accumulate_fragment(const char *frag_buf, size_t frag_len, 
			    PPM_frame_info& pfi, bool flushOrNot);

  virtual int process_whole_frame(const char *frame, size_t len)=0;

  //free *img_buffer because frame size varies.
  void free_img_buffer();

  //The accumulated img put here
  char* img_buffer;

 protected:
  //needed if reconstruct img_buffer
  //int curr_img_width, curr_img_height;

  //The starting point for negotiating the fragment size
  int MUS;

  //PPM_frame_info pfi;
  int line_accumulated;

  //Is this the first line of a new PPM frame
  bool first; 

  //data_start_line: the minimum line # of lines_send_per_time
  //data_end_line:   the maximum # of lines sent each time
  //img_start_line:  the starting line # in the so-far-accumulated img.
  int data_start_line,data_end_line,img_start_line;

  //flag array to indicate the whole frame has been accumulated
  Flag_Array line;

  int frame_height;
  int frame_width;
  unsigned long frame_no;
  unsigned long curr_PPM_frame_num;

  //for color manipulation
  int rpos, gpos, bpos;

  int bytes_per_line;
  int bits_per_pixel;

  //for X manipulation
  bool little_endian;
  
  //for debug
  int BytesReceived;
  //int PPMfile[2];
  int fileidx;
};

#endif
