/*******************************************************************************
 * FTimer.h 
 *
 * Falcon Timing Routines.
 * 
 *    This is a set of macros for timing based on the Portable Timing Routines 
 *    specification from the Parallel Tools Consortium.
 *    For an example of how to use them, take a look at Timing.c.
 *    The macros have been tested on Irix's, Sun's, and Solaris'.
 *    
 *    Remember to use define FTR_TIMING to include the timing macros
 *    (e.g. cc -DFTR_TIMING -o Timing Timing.c) 
 *
 */

#ifndef FTIMER_H
#define FTIMER_H

#include <sys/time.h>
#include <sys/resource.h>
#include <stdio.h>

typedef struct timeval FTR_timer_val;

static struct rusage usage;

typedef FTR_timer_val FTR_cpu_timer_val; /* opaque timer value for CPU time */
typedef FTR_timer_val FTR_usr_timer_val; /* opaque timer value for user CPU time */
typedef FTR_timer_val FTR_sys_timer_val; /* opaque timer value for system CPU time */
typedef FTR_timer_val FTR_wall_timer_val; /* opaque timer value for wallclock time */

int getrusage(int, struct rusage *);  

/*******************************************************************************
 * FTR_Get_cpu_time()
 *
 * Gets the opaque timer value for CPU time.
 *
 */
#ifdef FTR_TIMING

#define FTR_Get_cpu_time(reg) \
 if(getrusage(RUSAGE_SELF, &usage) == -1){perror("getrusage");}\
 reg.tv_sec = usage.ru_stime.tv_sec + usage.ru_utime.tv_sec; \
 reg.tv_usec = usage.ru_stime.tv_usec + usage.ru_utime.tv_usec;

#else

#define FTR_Get_cpu_time(reg) \
 reg.tv_sec = 0; \
 reg.tv_usec = 0;

#endif /* FTR_TIMING */

/* end FTR_Get_cpu_time() */


/*******************************************************************************
 * FTR_Get_usr_time()
 *
 * Gets the opaque timer value for user time.
 *
 */
#ifdef FTR_TIMING

#define FTR_Get_usr_time(reg) \
 if(getrusage(RUSAGE_SELF, &usage) == -1){perror("getrusage");}\
 reg.tv_sec = usage.ru_utime.tv_sec; \
 reg.tv_usec = usage.ru_utime.tv_usec;

#else

#define FTR_Get_usr_time(reg) \
 reg.tv_sec = 0; \
 reg.tv_usec = 0;

#endif /* FTR_TIMING */

/* end FTR_Get_usr_time() */


/*******************************************************************************
 * FTR_Get_sys_time()
 *
 * Gets the opaque timer value for system time.
 *
 */
#ifdef FTR_TIMING

#define FTR_Get_sys_time(reg) \
 if(getrusage(RUSAGE_SELF, &usage) == -1){perror("getrusage");}\
 reg.tv_sec = usage.ru_stime.tv_sec; \
 reg.tv_usec = usage.ru_stime.tv_usec;

#else

#define FTR_Get_sys_time(reg) \
 reg.tv_sec = 0; \
 reg.tv_usec = 0;

#endif /* FTR_TIMING */

/* end FTR_Get_sys_time() */


/*******************************************************************************
 * FTR_Get_wall_time()
 *
 * Gets the opaque timer value for wallclock time.
 *
 */
#ifdef FTR_TIMING

#define FTR_Get_wall_time(reg) \
 gettimeofday(&reg, NULL);

#else

#define FTR_Get_wall_time(reg) \
 reg.tv_sec = 0; \
 reg.tv_usec = 0;

#endif /* FTR_TIMING */

/* end FTR_Get_wall_time() */


/*******************************************************************************
 * FTR_Get_cpu_interval()
 *
 * Calculates the interval between two opaque values for CPU time.
 *
 */
#ifdef FTR_TIMING

#define FTR_Get_cpu_interval(start, end, inter) \
 inter.tv_sec = end.tv_sec - start.tv_sec; \
 inter.tv_usec = end.tv_usec - start.tv_usec; 

#else

#define FTR_Get_cpu_interval(start, end, inter) \
 inter.tv_sec = 0; \
 inter.tv_usec = 0;

#endif /* FTR_TIMING */

/* end FTR_Get_cpu_interval() */


/*******************************************************************************
 * FTR_Get_usr_interval()
 *
 * Calculates the interval between two opaque values for user time.
 *
 */
#ifdef FTR_TIMING

#define FTR_Get_usr_interval(start, end, inter) \
 inter.tv_sec = end.tv_sec - start.tv_sec; \
 inter.tv_usec = end.tv_usec - start.tv_usec;

#else

#define FTR_Get_usr_interval(start, end, inter) \
 inter.tv_sec = 0; \
 inter.tv_usec = 0;

#endif /* FTR_TIMING */

/* end FTR_Get_usr_interval() */


/*******************************************************************************
 * FTR_Get_sys_interval()
 *
 * Calculates the interval between two opaque values for system time.
 *
 */
#ifdef FTR_TIMING

#define FTR_Get_sys_interval(start, end, inter) \
 inter.tv_sec = end.tv_sec - start.tv_sec; \
 inter.tv_usec = end.tv_usec - start.tv_usec;

#else

#define FTR_Get_sys_interval(start, end, inter) \
 inter.tv_sec = 0; \
 inter.tv_usec = 0;

#endif /* FTR_TIMING */

/* end FTR_Get_sys_interval() */


/*******************************************************************************
 * FTR_Get_wall_interval()
 *
 * Calculates the interval between two opaque values for wallclock time.
 *
 */
#ifdef FTR_TIMING

#define FTR_Get_wall_interval(start, end, inter) \
 inter.tv_sec = end.tv_sec - start.tv_sec; \
 inter.tv_usec = end.tv_usec - start.tv_usec;

#else

#define FTR_Get_wall_interval(start, end, inter) \
 inter.tv_sec = 0; \
 inter.tv_usec = 0;

#endif /* FTR_TIMING */

/* end FTR_Get_wall_interval() */


/*******************************************************************************
 * FTR_Seconds()
 *
 * Gets the seconds of a opaque timer as a double.
 *
 */
#ifdef FTR_TIMING

#define FTR_Seconds(x) \
 ((double)((x).tv_sec) + (1.0e-6 * (x).tv_usec))

#else 

#define FTR_Seconds(x) \
 ((double)(0))

#endif /* FTR_TIMING */

/* end FTR_Seconds() */


/*******************************************************************************
 * FTR_Zero_timer()
 *
 * Zeros the opaque timer.
 *
 */
#define FTR_Zero_timer(reg) \
 reg.tv_sec = 0; reg.tv_usec = 0;

/* end FTR_Zero_timer() */

#endif /* FTIMER_H */


/* end FTimer.h ***************************************************************/
