/* -*-C-*-
*******************************************************************************
*
* File:         list.c
* RCS:          $Id: list.c,v 1.1.1.1 2001/07/03 20:30:44 fabianb Exp $
* Description:  Basic list lib
* Author:       Fabian Ernesto Bustamante
*               Systems Research Group
*               College of Computing
*               Georgia Institute of Technology
* Created:      Wed May 09, 2001 at 15:54:44
* Modified:     Wed May 09, 2001 at 15:55:11 fabianb@cc.gatech.edu
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
* (C) Copyright 2001, Georgia Institute of Technology, all rights reserved.
*
*******************************************************************************
*/

#if HAVE_CONFIG_H
#  include <src/config.h>
#endif

#include <stdarg.h>
#include <stddef.h>
#include <stdlib.h>
#include <strings.h>
#include "list.h"

struct _list_node {
  void *element;		/* content */
  struct _list_node* next;
  struct _list_node* prev;
};


/* 
 * A double linked list with sentinel. "match" is not used by linked list but 
 * by datatypes that will be derived later from them. We assume strcmp 
 * semantics.
 *
 */

struct List_T {
  int size;
  int (*match) (void*, void*);
  /* head of list is sentinel->next and tail is sentinel->prev */
  struct _list_node* sentinel;
  struct _list_node* current;
};


/*----------------------------------------------------------------------------*/

/*
 * List_length: Returns number of elements in list
 * 
 * Parameters: List handler
 * 
 * Returns: Number of elements in list
 *
 */

int
List_length(List_T list) 
{
  if (!list) return -1; 
  else return list->size;
} /* end List_length */


/*----------------------------------------------------------------------------*/

/*
 * List_new: Creates and initialize a new list
 * 
 * Parameters: -
 * 
 * Returns: New list handler
 *
 */

List_T 
List_new(void)
{
  List_T list;
  
  list = (List_T) malloc(sizeof(*list));
  if (list == NULL) return NULL;
  bzero((void*) (list), sizeof(*list));

  /* get and insert sentinel */
  list->sentinel = list->current = 
    (struct _list_node *) malloc(sizeof(struct _list_node));
  if (list->sentinel == NULL) {
    free(list);
    return NULL;
  }

  /* set sentinel's pointers and list's size for an empty list */
  list->sentinel->next = list->sentinel->prev = list->sentinel;
  list->size = 0;
  return list;
} /* end List_new() */


/*----------------------------------------------------------------------------*/

/*
 * List_list: Creates a list from a variable argument list
 *  
 * Parameters: Pointer to variable argument list
 * 
 * Returns: New list handler
 *
 * NOTE: Potential problem of portability
 *
 */

List_T 
List_list(void *x, ...) {
  va_list ap;
  List_T list;

  list = List_new();  
  va_start(ap, x);
  for ( ; x; x = va_arg(ap, void *)) 
    List_append(list, x);
  va_end(ap);
  return list;
} /* end List_list */


/*----------------------------------------------------------------------------*/

/*
 * List_delete: Removes the current element of the list and returns its 
 *   content
 * 
 * Parameters: List handler
 * 
 * Returns: Content of the remove element
 *
 */

void * 
List_delete(List_T list)
{
  struct _list_node *tmp;
  void * aux;
 
  if (list->size != 0) {
    list->current->prev->next = list->current->next;
    list->current->next->prev = list->current->prev;
    tmp = list->current;
    list->current = list->current->prev;
    list->size -= 1;
    aux = tmp->element;
    free(tmp);
    return aux;
  }

  return NULL;
} /* end List_delete() */


/*----------------------------------------------------------------------------*/

/*
 * List_free: Destructor for the list
 * 
 * Parameters: List handler
 * 
 * Returns: -
 *
 */

void 
List_free(List_T list)
{
  if (!list) return;
  while (list->size != 0)
    List_delete(list);

  free(list->sentinel);
  free(list);
} /* end List_free() */


/*----------------------------------------------------------------------------*/

/*
 * List_append: Inserts a new element in the list, placing it after the 
 *   current one, with el as its content
 * 
 * Parameters: List handler and new element's content
 * 
 * Returns: -
 *
 */

void 
List_append(List_T list, void* el)
{
  struct _list_node *tmp;

  tmp = (struct _list_node *) malloc(sizeof(struct _list_node));
  if (tmp == NULL) return;
  tmp->element = el;

  tmp->prev = list->current;
  tmp->next = list->current->next;
  tmp->next->prev = tmp;
  list->current->next = tmp;
  list->current = tmp;
  list->size += 1;
} /* end List_append() */


/*----------------------------------------------------------------------------*/

/*
 * List_head: Makes current the head of the list
 * 
 * Parameters: List handler
 * 
 * Returns: Pointer to new current node
 *
 */

void * 
List_head(List_T list)
{
  if (list->size != 0) {
    list->current = list->sentinel->next;
    return list->current->element;
  }
  return NULL;
} /* end List_head() */


/*----------------------------------------------------------------------------*/

/*
 * List_tail: Makes current the tail of the list
 * 
 * Parameters: List handler
 * 
 * Returns: Pointer to new current node
 *
 */

void * 
List_tail(List_T list)
{
  if (list->size != 0) {
    list->current = list->sentinel->prev;
    return list->current->element;
  }
  return NULL;
} /* end List_tail() */


/*----------------------------------------------------------------------------*/

/*
 * List_current: Returns pointer to current node
 * 
 * Parameters: List handler
 * 
 * Returns: Pointer to current node
 *
 */

void * 
List_current(List_T list)
{
  if (list->size != 0)
    return list->current->element;
  return NULL;
} /* end List_current() */


/*----------------------------------------------------------------------------*/

/*
 * List_next: Moves current node to next element
 * 
 * Parameters: List handler
 * 
 * Returns: Pointer to new current node
 *
 */

void * 
List_next(List_T list)
{
  if (list->size != 0) {
    if (list->current->next == list->sentinel)
      return NULL;		/* current is the tail */
    list->current = list->current->next;
    return list->current->element;
  }
  return NULL;
} /* end List_next() */


/*----------------------------------------------------------------------------*/

/*
 * List_previous: Moves current node to previous element
 * 
 * Parameters: List handler
 * 
 * Returns: Pointer to new current node
 *
 */

void * 
List_previous(List_T list)
{
  if (list->size != 0) {
    if (list->current->prev == list->sentinel)
      return NULL;		/* current is the head */
    list->current = list->current->prev;
    return list->current->element;
  }
  return NULL;
} /* end List_previous() */


/*----------------------------------------------------------------------------*/

/*
 * List_search: Searchs list for element that correctly compares to the given 
 *   one
 * 
 * Parameters: List handler, a guiding element, and the comparation function
 * 
 * Returns: The first element that successfuly compares or null
 *
 */

void * 
List_search(List_T list, void* el, int (*comp) (void *, void *))
{
  if (list->size != 0) {
    list->current = list->sentinel->next;
    
    /* search for the node with eq element, assuming strcmp() behaviour */
    while((list->current != list->sentinel->prev) && 
	  (comp(el, list->current->element) != 0))
      list->current=list->current->next;
    
    if (!comp(el, list->current->element))
      return list->current->element;

    return NULL;
  }
  return NULL;
} /* end List_search() */


/*----------------------------------------------------------------------------*/

/*
 * List_map: Applies 'apply' to all elements in list
 * 
 * Parameters: List handler and function to apply
 * 
 * Returns: -
 *
 */

void 
List_map(List_T list, void apply(void **x, void *cl), void *cl)
{
  if (apply == NULL) return;
  
  if (list->size != 0) {
    list->current = list->sentinel->next;
    while (list->current != list->sentinel->prev) {
      apply(&list->current->element, cl);
      list->current = list->current->next;
    } 
    apply(&list->current->element, cl);
  }
} /* end List_map() */


/*----------------------------------------------------------------------------*/

/*
 * List_beforeInsert:  Insert a new element before the current one
 * 
 * Parameters: List handler and the content of the new element
 * 
 * Returns: -
 *
 */

void 
List_beforeInsert(List_T list, void* el)
{
  struct _list_node * tmp;

  /* if the list is empty (current == sentinel), inserts after header */
  if (list->size == 0) {
    List_append(list, el);
    return;
  }

  tmp = (struct _list_node *)malloc(sizeof(struct _list_node));
  if (tmp == NULL) return;
  tmp->element = el;

  tmp->next = list->current;
  tmp->prev = list->current->prev;
  tmp->prev->next = tmp;
  list->current->prev = tmp;
  list->current = tmp;
  list->size += 1;
} /* end List_beforeInsert() */


/*----------------------------------------------------------------------------*/

/*
 * List_orderedInsert: Insert a new element in order
 * 
 * Parameters: List handler, content of the new element, and comp function
 *   to determine order
 * 
 * Returns: -
 *
 */

void 
List_orderedInsert(List_T list, void* el, int (*comp)(void *, void *))
{  
  if (list->size != 0) {
    list->current = list->sentinel->next;
    while((list->current != list->sentinel->prev) && 
	  (comp(el, list->current->element) > 0))
      list->current = list->current->next;
    if (comp(el, list->current->element) < 0)
      List_beforeInsert(list, el);
    else
      List_append(list, el);
  } else
    List_append(list, el);
} /* end List_orderedInsert() */


/*----------------------------------------------------------------------------*/

/*
 * List_sort: Sorts the list, according to 'comp' (strcmp's semantic), into 
 *   a new list
 * 
 * Parameters: List handler and comp function
 * 
 * Returns: New, ordered, list
 *
 */

List_T
List_sort(List_T list, int (*comp)(void *, void *))
{
  List_T tmp;
  void *el;

  tmp = List_new();

  if (list->size != 0) {
    list->current = list->sentinel->next;
    while (list->current != list->sentinel->prev) {
      el = list->current->element;
      List_orderedInsert(tmp, el, comp);
      list->current = list->current->next;
    }  
    el = list->current->element;
    List_orderedInsert(tmp, el, comp);
  }
  return tmp;
} /* end List_sort() */


/*----------------------------------------------------------------------------*/

/*
 * List_toArray: Makes an array out of a list, placing 'end' as the last 
 *   element
 * 
 * Parameters: List handler and last element for the array
 * 
 * Returns: Pointer to the array
 *
 */

void **
List_toArray(List_T list, void *end) {
  int i, n = List_length(list);
  void **array = malloc((n + 1)*sizeof (*array));

  if (array == NULL) return NULL;

  List_head(list);
  for (i = 0; i < n; i++) {
    array[i] = List_current(list);
    List_next(list);
  }
  array[i] = end;
  return array;
} /* end List_toArray */

/*----------------------------------------------------------------------------*/
