/* -*-C-*-
 * @ Wenchang Yan
 * @date Mar, 2002
 * 
*******************************************************************************
*
* File:         vdev.c
* RCS:          $Id: vdev.c,v 1.1.1.1 2001/07/03 20:30:44 fabianb Exp $
* Description:  Library to access USB video cams in Linux (RH 7.1, Kernel 2.4) 
*               Look at the included README file.
* Author:       Fabian E. Bustamante
*               Systems Research Group
*               College of Computing
*               Georgia Institute of Technology
* Created:      Sun May 13, 2001 at 10:27:18
* Modified:     Mon Jul 02, 2001 at 15:57:06 fabianb@cc.gatech.edu
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
* (C) Copyright 2001, Georgia Institute of Technology, all rights reserved.
*
*******************************************************************************
*/

#if HAVE_CONFIG_H
#  include <src/config.h>
#endif

#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <unistd.h>
#include <sys/mman.h>
#include <linux/videodev.h>

#include <vdev.h>
#include <pwc-ioctl.h>
#include <list.h>
#include <ppm.h>

/*vvv*/
#undef DEBUG
#define DEBUG 0
/*^^^*/

#ifdef VDEV_DEBUG
#  define PRINT_DEBUG(x) printf x ;
#else 
#  define PRINT_DEBUG(x)
#endif

#define VDEV_SUCCESS  (0)
#define VDEV_FAILURE  (-1)



typedef enum {false, true} boolean;


struct V_VDev_T {
  char nodeName[VDEV_NAMELEN];	/* devname, e.g. /dev/video0 */
  char intfName[VDEV_NAMELEN];	/* V4L device name */
  int camFd;			/* cam file descriptor */
  struct video_picture vpic;
  struct video_mbuf vmbuf;	/* size + frames + offsets[] */
  int buffers;
  int curBuffer;
  boolean validated;
  boolean hasFramerate;
  int image_w, image_h;
  u_char* vid_buffer;
  u_char* ppm_buffer;
  u_char* tmp_buffer;		/* for deep copy */
  int vid_buffer_size, image_buffer_size;
  int max_w, max_h;		/* Max and min widht and height */
  int min_w, min_h;
  int framerate;
  int sizeRead;			/* Size of last read */
  int tickCounter;		/* need if !hasFramerate */
  boolean imageReady;
  int palette;
};

List_T Devices = NULL;
VDev_T *FrameDevices = NULL; /* Array of devices we control framerate for */
int FrameDLen = 0;
int MaxFramerate;



/*----------------------------------------------------------------------------*/

/*
 * TickOn: Update tick counters and readImage for those who need it
 *   Since we have only one alarm we set it to a fix rate, set framerates to 
 *   multipliers of it, and keep a counter of ticks to know when the period
 *   for a frame update has passed.
 *
 * Parameters: signo - signal number
 * 
 * Returns: -
 *
 */

static RETSIGTYPE
TickOn (int signo)
{
  int i, maxTickCounter;

  for (i = 0; i < FrameDLen; ++i) {
    if (FrameDevices[i]->camFd > 0) { /* could have been close since */
      ++(FrameDevices[i]->tickCounter);
      
      maxTickCounter = MaxFramerate/FrameDevices[i]->framerate;
      if (FrameDevices[i]->tickCounter == maxTickCounter) {
	FrameDevices[i]->imageReady = true;
	FrameDevices[i]->tickCounter = 0;
      }
    }
  }

} /* end TickOn */


/*----------------------------------------------------------------------------*/

/*
 * UpdateFrameDevices: Create the new array of pointer to the devices we must
 *   take care of (those opened and w/o framerate setting). Set the timer
 *   to the mcd of minimum-period/fps and adjust tick-counters of previously
 *   open devices under control. Reset the timer.
 *
 * Parameters: -
 *
 * Returns: Zero on success, or -1 if an error occurred.
 * 
 */

static int
UpdateFrameDevices ()
{
  int i, j, n = List_length(Devices);
  VDev_T node;
  struct itimerval itv;
  long tickPeriod;

  free(FrameDevices);

  FrameDLen = 0;
  List_head(Devices);
  for (i = 0; i < n; ++i) {
    node = List_current(Devices);
    if (node->camFd > 0 && node->hasFramerate == false) 
      ++FrameDLen;
    List_next(Devices);
  }

  FrameDevices = malloc (FrameDLen * sizeof(*FrameDevices));
  j = 0;			/* index into FrameDevices */
  List_head(Devices);
  for (i = 0; i < n; ++i) {
    node = List_current(Devices);
    if (node->camFd > 0 && node->hasFramerate == false) 
      FrameDevices[j] = node;
    List_next(Devices);
  }

  /* Set timer to tick for max framerate periodicity (a way around mcd) */
  MaxFramerate = 40;
  for (j = 0; j < FrameDLen; ++j) {
    if (FrameDevices[j]->framerate > MaxFramerate) 
      MaxFramerate = FrameDevices[j]->framerate;
    /* Coarse but the simplest way to go */
    FrameDevices[j]->tickCounter = 0;
  }
  
  if (signal(SIGALRM, TickOn) == SIG_ERR) return VDEV_FAILURE;
  tickPeriod = 1000.0/MaxFramerate;
  itv.it_value.tv_sec = tickPeriod*1e-3;
  itv.it_value.tv_usec = (tickPeriod - (itv.it_value.tv_sec*1e3))*1e3;
  /* if (setitimer(ITIMER_REAL, &itv, NULL)) return VDEV_FAILURE; */
  
  return VDEV_SUCCESS;
} /* end UpdateFrameDevices */


/*----------------------------------------------------------------------------*/

/*
 * MCapture: Do ioctl(MCAPTURE), start grabbing frame in buffer.
 *   This function will start the capture of an image into buffer 'buf' with
 *   the current width & height
 *
 * Parameters: vdev - handler for video device
 *             buf - the mmap buffer (will be set modulo total buffers)
 *
 * Returns: Zero on success, or -1 if an error occurred.
 * 
 */

static int 
MCapture(VDev_T vdev, int buf)
{
  struct video_mmap vm;		/* frame + height + width + format */

# if DEBUG
  fprintf (stderr,"MCapture buffer %d\n", buf);
# endif

  if (vdev == NULL) 
    return VDEV_FAILURE;

  if (vdev->vmbuf.size == 0)
    return VDEV_SUCCESS;
  
  vm.frame = buf % vdev->buffers;
  vm.format = vdev->palette;
  vm.width = vdev->image_w;
  vm.height = vdev->image_h;
   
# if DEBUG
  fprintf(stderr," [MCapture(): video_mmap: frame=%d, format=%d, width=%d, height=%d\n",vm.frame,vm.format,vm.width,vm.height);
  fprintf(stderr," [MCapture(): video_mbuf: size=%d, frames=%d, offsets[0]=%d\n",vdev->vmbuf.size,vdev->vmbuf.frames,vdev->vmbuf.offsets[0]);
# endif

  if (ioctl(vdev->camFd, VIDIOCMCAPTURE, &vm) < 0)
    return VDEV_FAILURE;

  return VDEV_SUCCESS;
} /* end MCapture */


/*----------------------------------------------------------------------------*/

/*
 * MSync: Do ioctl(SYNC), releasing buffer
 * 
 * Parameters: vdev - handler for video device
 * 
 * Returns: Zero on success, or -1 if an error occurred.
 *
 */

static int 
MSync(VDev_T vdev)
{
  PRINT_DEBUG (("MSync\n"));

  if (vdev == NULL) 
    return VDEV_FAILURE;
  
  if (vdev->vmbuf.size == 0)
    return VDEV_SUCCESS;
  
  if (ioctl(vdev->camFd, VIDIOCSYNC, &(vdev->curBuffer)) < 0)
    return VDEV_FAILURE;

  return VDEV_SUCCESS;
} /* end MSync */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_init: Scan available devices, build a list of them (Devices), ...
 *            
 * Parameters: -
 * 
 * Returns: Zero on success, or -1 if an error occurred.
 *
 */
/*vvv*/
/*
extern int 
VDev_init()
*/
extern int VDev_init(int w,int h)
{
  int i;
  char devname[VDEV_NAMELEN];
  struct stat devstat;
  VDev_T v;

  PRINT_DEBUG (("VDev_init\n"));

  if (Devices != NULL) return 0; /* Already initialized */

  Devices = List_new();
  if (Devices == NULL) return VDEV_FAILURE; /* Memory exhausted */

  for (i = 0; i < VDEV_MAXNUMDEV; i++) {
    sprintf(devname, "/dev/video%d", i);
    if (stat(devname, &devstat) == 0) {
      PRINT_DEBUG (("Testing %s\n", devname));

      if (!S_ISLNK(devstat.st_mode)) {
	/*vvv*/
	/*v = VDev_new(devname);*/
	v = VDev_new(devname,w,h);
	/*^^^*/

	if (!VDev_isValid(v)) {
	  VDev_free(v);
	} else {
	  List_append(Devices, (void*) v);
       	  PRINT_DEBUG (("This one is fine %s\n", devname));
	}
      }
    }
  }

  return VDEV_SUCCESS;
} /* end VDev_init */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_finalize: Clean up after us, including list of devices
 *
 * Parameters: -
 * 
 * Returns: -
 *
 */

extern void
VDev_finalize()
{
  int size, i;
  VDev_T node;

  PRINT_DEBUG (("VDev_finalize\n"));

  size = (int) List_length(Devices);
  for (i = 0; i < size; i++) {
    node = List_current(Devices);
    VDev_free(node);
    List_delete(Devices);
    List_next(Devices);
  }
  List_free(Devices);

  FrameDLen = 0;
  free(FrameDevices);
} /* end VDev_finalize */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_numVDevices: Returns number of video devices
 *
 * Parameters: -
 * 
 * Returns: Number of video devices, or -1 if an error occurred.
 *
 */

int 
VDev_numVDevices ()
{
  PRINT_DEBUG (("VDev_numVDevices\n"));

  return List_length(Devices);	/* List_length takes care of null list */
} /* end VDev_numVDevices */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_getVDevice: Returns a pointer to the indicated video device
 *
 * Parameters: Number of device
 * 
 * Returns: Handler for video device, or NULL if an error occurred.
 *
 */

VDev_T
VDev_getVDevice(int n)
{
  int i, size;
  VDev_T v;

  PRINT_DEBUG (("VDev_getDevice\n"));

  if (!Devices) return NULL;
  size = List_length(Devices);
  if (n > size) return NULL;
  v = (VDev_T) List_head(Devices);
  for (i = 1; i < n; i++)
    v = (VDev_T) List_next(Devices);
  return v;
} /* end VDev_getVDevice */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_new: Constructor for video device. It opens the devices only to get 
 *   configuration information and validate it. The user must explicitly 
 *   open it through VDev_open().
 *
 * Parameters: devname
 * 
 * Returns: Pointer to new video device on success, or NULL if an error 
 *   occurred.
 *
 */
/*vvv*/
/*
VDev_T 
VDev_new(char* devname)
*/
VDev_T VDev_new(char* devname, int w, int h)
{
  int camFd;
  VDev_T vdev;
  struct video_capability vcap;
  struct video_window vwin;
  struct video_picture vpic;

  boolean validated = false;
  /*vvv*/
  //boolean hasFramerate = false;
  boolean hasFramerate = true;
  struct video_buffer videobuffer;
  /*^^^*/
  int image_w = 0, image_h = 0;

  PRINT_DEBUG (("VDev_new\n"));

  camFd = open(devname, O_RDONLY);
  if (camFd < 0) {    
    if (errno == EBUSY)
      validated = true;
  } else {
    validated = true;
    if (ioctl(camFd, VIDIOCGCAP, &vcap) < 0) { 
      return NULL;
    }
    if (ioctl(camFd, VIDIOCGWIN, &vwin) == 0) {
      image_w = vwin.width;
      image_h = vwin.height;

#     if DEBUG
      fprintf(stderr,"[VDev_new(): video_window: %dx%d\n",image_w,image_h);
#     endif

#if defined(PWS_FPS_SHIFT)
      if ((vwin.flags & PWC_FPS_FRMASK) >> PWC_FPS_SHIFT) {
	hasFramerate = true;
	PRINT_DEBUG (("Video device supports framerate setting [%x].\n", 
		      vwin.flags));
      }
#endif
    }
    
    if (ioctl(camFd, VIDIOCGPICT, &vpic) <0) {
      perror("VIDIOCGPICT");
      return NULL;
    }
  }

  /* set the properties */
# if DEBUG
  fprintf(stderr, "[VDev_new(): video_capture: type=%d, channels=%d, max=%dx%d, min=%dx%d\n",vcap.type,vcap.channels,vcap.maxwidth,vcap.maxheight,vcap.minwidth,vcap.minheight);
  fprintf(stderr, "[VDev_new(): video_picture: bright=%d, hue=%d, color=%d, contrast=%d, whiteness=%d, depth=%d, palette=%d\n",vpic.brightness,vpic.hue,vpic.colour,vpic.contrast,vpic.whiteness,vpic.depth,vpic.palette);
# endif

  if (vcap.type & VID_TYPE_MONOCHROME) {
    vpic.depth=8;
    vpic.palette=VIDEO_PALETTE_GREY; 
    if(ioctl(camFd, VIDIOCSPICT, &vpic) < 0) {
      vpic.depth=6;
      if(ioctl(camFd, VIDIOCSPICT, &vpic) < 0) {
        vpic.depth=4;
        if(ioctl(camFd, VIDIOCSPICT, &vpic) < 0) {
#if DEBUG
          fprintf(stderr, "Unable to find a supported capture format.\n");
#endif
          close(camFd);
          return NULL;
        }
      }
    }
  } else {
    vpic.depth=24;
    vpic.palette=VIDEO_PALETTE_RGB24;
    
    if(ioctl(camFd, VIDIOCSPICT, &vpic) < 0) {
      vpic.depth=16;
      vpic.palette=VIDEO_PALETTE_RGB565;
      
      if(ioctl(camFd, VIDIOCSPICT, &vpic)==-1) {
        vpic.depth=15;
        vpic.palette=VIDEO_PALETTE_RGB555;
        
        if(ioctl(camFd, VIDIOCSPICT, &vpic)==-1) {
#         if DEBUG
          fprintf(stderr, "Unable to find a supported picture format.\n");
#         endif
          return NULL;
        }
      }
    }
  }

# if DEBUG
  fprintf(stderr, "[VDev_new(): after set video_picture:\n");
  fprintf(stderr, "[VDev_new(): video_picture: bright=%d, hue=%d, color=%d, contrast=%d, whiteness=%d, depth=%d, palette=%d\n",vpic.brightness,vpic.hue,vpic.colour,vpic.contrast,vpic.whiteness,vpic.depth,vpic.palette);
# endif
  
  /* Get video_buffer */
# if DEBUG
  if (ioctl(camFd,VIDIOCGFBUF, &videobuffer)==0) {
    fprintf(stderr," [VDev_new(): video_buffer: width=%d, height=%d, depth=%d, bytesperline=%d\n", videobuffer.width,videobuffer.height,videobuffer.depth,videobuffer.bytesperline);
  }
# endif

  /* All is fine; create the entry for this dev and initialize it */
  vdev = (VDev_T) malloc(sizeof(*vdev));
  if (vdev == NULL) return NULL;
  bzero((void*) (vdev), sizeof(*vdev));
  
  strncpy(vdev->nodeName, devname, VDEV_NAMELEN);
  vdev->nodeName[VDEV_NAMELEN] = '\0';
  if (ioctl(camFd, VIDIOCGPICT, &(vdev->vpic)) < 0) {
    free(vdev);
    return NULL;
  }

  //vdev->palette = vdev->vpic.palette;
  vdev->vpic = vpic;

  vdev->camFd = -1;
  vdev->validated = validated;
  vdev->hasFramerate = hasFramerate;
  strncpy(vdev->intfName, vcap.name, VDEV_NAMELEN);
  vdev->intfName[VDEV_NAMELEN] = '\0';
  vdev->buffers = 0;
  vdev->curBuffer = 0;
  /*vvv*/
  vdev->image_w = w;/*image_w;*/
  vdev->image_h = h;/*image_h;*/
  vdev->palette = vpic.palette;
  /*^^^*/
  vdev->image_buffer_size = vdev->image_w * vdev->image_h * 3; /* 3 for 4 */
  vdev->vid_buffer_size = vdev->image_buffer_size;

  vdev->vmbuf.size = 0;
  vdev->vmbuf.frames = 0;
  if (vdev->hasFramerate == false) vdev->tickCounter = 0.0;
  else vdev->tickCounter = -1.0;
  vdev->max_w = vcap.maxwidth;
  vdev->max_h = vcap.maxheight;
  vdev->min_w = vcap.minwidth;
  vdev->min_h = vcap.minheight;
  vdev->sizeRead = -1;
  if (hasFramerate) vdev->framerate = 25;
  else vdev->framerate = 20;	/* if it is up to us - see VDev_setFramerate */
  vdev->imageReady = true;

# if DEBUG
  fprintf(stderr," [VDev_new(): vid_buffer_size=image_buffer_size=%d\n",vdev->vid_buffer_size);
  fprintf(stderr," [VDev_new(): max_w=%d, max_h=%d, min_w=%d, min_h=%d\n",vdev->max_w,vdev->max_h,vdev->min_w,vdev->min_h);
# endif

  close(camFd);

  return vdev;
} /* end VDev_new */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_free: Destructor for video devices
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: -
 *
 */

extern void
VDev_free(VDev_T vdev)
{
  PRINT_DEBUG (("VDev_free\n"));

  if (!vdev) return;
  if (vdev->vmbuf.size > 0) {
    MSync(vdev);
    munmap(vdev->vid_buffer, vdev->vid_buffer_size);

  } else 
    free(vdev->vid_buffer);

  free(vdev);
  return;
} /* end VDev_free */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_isValid: Check it is a valid video device.
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: Zero on success, or -1 if an error occurred.
 *
 */

extern int
VDev_isValid(VDev_T vdev)
{
  PRINT_DEBUG (("VDev_isValid\n"));

  if (vdev != NULL)
    return (int) vdev->validated;
  return (int) false;
} /* end VDev_isValid */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_open: Open associated file
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: Zero on success, or -1 if an error occurred.
 *
 */

extern int
VDev_open(VDev_T vdev, int buf)
{
  PRINT_DEBUG (("VDev_open\n"));

  if (vdev == NULL) return VDEV_FAILURE;

  vdev->camFd = open(vdev->nodeName, O_RDONLY);
  if (vdev->camFd < 0) {
    return VDEV_FAILURE;
  }
  
  vdev->buffers = buf;
  if (buf < 1) vdev->buffers = 1;
  if (buf > 8) vdev->buffers = 8;
  

  PRINT_DEBUG((" [VDev_open(): camFd=%d\n",vdev->camFd));

  /* see if we can use mmap */
  if (ioctl(vdev->camFd, VIDIOCGMBUF, &(vdev->vmbuf)) == 0) {
    vdev->vid_buffer_size = vdev->vmbuf.size;
    if (vdev->vmbuf.frames < vdev->buffers) {
      perror("Not enough buffers from mbuf.");
      vdev->vmbuf.size = 0;
    }
    vdev->vid_buffer = (u_char*) mmap(NULL, vdev->vid_buffer_size, 
				      PROT_READ, MAP_SHARED, 
				      vdev->camFd, 0);
    if (vdev->vid_buffer == (u_char*) -1) {
      perror("Mmap failed.");
      vdev->vmbuf.size = 0;
    }
  }

  if (vdev->vmbuf.size == 0) {
    vdev->vid_buffer_size = vdev->image_buffer_size;
    vdev->vid_buffer = (u_char*) malloc(vdev->vid_buffer_size);
    if (!vdev->vid_buffer) {
      perror("Not enough memory at all");
      free(vdev);
      return VDEV_FAILURE;
    }
  }
  
  if (MCapture(vdev, vdev->curBuffer) < 0) {
    perror("Error starting mcapture");
    return VDEV_FAILURE;
  }

  if (vdev->hasFramerate == false) UpdateFrameDevices();

  return VDEV_SUCCESS;
} /* end VDev_open */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_close: Close associated file
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: Zero on success, or -1 if an error occurred.
 *
 */

extern int
VDev_close(VDev_T vdev)
{
  PRINT_DEBUG (("VDev_close\n"));

  if (vdev != NULL) {
    close(vdev->camFd);
    vdev->camFd = -1;
    return VDEV_SUCCESS;
  }

  return VDEV_FAILURE;
} /* end VDev_close */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_readImage:
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: Size read in bytes on success, or -1 if an error occurred.
 *
 */

extern int
VDev_readImage(VDev_T vdev)
{
  u_char *src;
  int size;

  PRINT_DEBUG (("VDev_readImage\n"));

  if ((vdev == NULL) || (vdev->camFd < 0)) return VDEV_FAILURE;

  if (vdev->vmbuf.size > 0) {
    if (MSync(vdev) < 0) return VDEV_FAILURE;
    src = vdev->vid_buffer + vdev->vmbuf.offsets[vdev->curBuffer];
    if (vdev->buffers > 1) {
      if (MCapture(vdev, vdev->curBuffer + 1) < 0) {
	perror("Error starting mcapture");
	return VDEV_FAILURE;
      }
    }
    size = vdev->image_buffer_size * sizeof(u_char);

    PRINT_DEBUG((" [VDev_readImage(): (vdev->vmbuf.size>0)==TRUE, size=%d\n",size));

  } else {

    if (vdev->vid_buffer == NULL) return -1;
    if ((size = read(vdev->camFd, vdev->vid_buffer,
		     vdev->image_buffer_size)) < 0)
      return VDEV_FAILURE;
    src = vdev->vid_buffer;

    PRINT_DEBUG((" [VDev_readImage(): (vdev->vmbuf.size>0)==FALSE, size=%d, vid_buffer=%x\n",size,vdev->vid_buffer));
  }

  vdev->curBuffer = (vdev->curBuffer + 1) % vdev->buffers;
  if (vdev->buffers == 1) {
    if (MCapture(vdev, vdev->curBuffer) < 0) {
      perror("Error starting mcapture");
      return VDEV_FAILURE;
    }
    PRINT_DEBUG((" [VDev_readImage(): (vdev->buffers==1)==TRUE\n"));
  }

  if (src == NULL) {
    perror("Problems reading image");
    return VDEV_FAILURE;
  }

  /* do a deep copy for now */
  free(vdev->tmp_buffer);
  if (!(vdev->tmp_buffer = (u_char*) malloc(size * sizeof(u_char)))) {
    perror("Out of memory");
    return VDEV_FAILURE;
  }

  memcpy(vdev->tmp_buffer, src, size);

  PRINT_DEBUG((" [VDev_readImage(): %d bytes have been memcpied\n",size));
  return size;
} /* end VDev_readImage */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_getImage: Retrieve pointer to an RGB image; note that this is a 
 *    pointer, not a (shallow) copy. The image is a 32 bit deep image.
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: Pointer to an RGB image on success, or NULL if an error 
 *   occurred.
 *
 */

extern u_char* 
VDev_getImage(VDev_T vdev)
{
  int sizeRead;

  PRINT_DEBUG (("VDev_getImage\n"));

  if (vdev == NULL) return NULL;
  
  vdev->imageReady = true; /* ### For now - force it, we should be using fps */
  if (vdev->imageReady) {
    sizeRead = VDev_readImage(vdev);
    if (sizeRead < 0)
      perror("Problems loading image");
    vdev->sizeRead = sizeRead;
    vdev->imageReady = false;
  } else {
    PRINT_DEBUG (("VDev_getImage - using pre-read image.\n"));
  }
  
  if (vdev->vid_buffer != NULL) 
    return vdev->tmp_buffer;	/* ### Notice I am returning the copy here */

  return NULL;
} /* end VDev_getImage */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_getPPMImage: Returns pointer to a PPM image; note that this is a 
 *    pointer, not a (shallow) copy.
 *    This is stupid, the only thing we are doing is switching around 
 *    some bytes!
 *    That's because we used PPM_MAXMAXVAL 255
 *
 * Parameters: vdev - handler for video device
 *             format - in [RGB, RBG, GRB, GBR, BGR, BRG]
 * 
 * Returns: Pointer to an PPM image on success, or NULL if an error 
 *   occurred.
 *
 */
extern u_char*
VDev_getPPMImage(VDev_T vdev, int format)
{
  u_char *pixarray;
  int row, col;
  int idx;
  int sizeRead;
  u_char val1, val2, val3;

  PRINT_DEBUG (("VDev_getPPMImage\n"));
  
  if (vdev == NULL || format < 1 || format > 6) return NULL;
  
  vdev->imageReady = true; /* ### For now - force it, we should be using fps */
  if (vdev->imageReady) {
    sizeRead = VDev_readImage(vdev);
    if (sizeRead < 0)
      perror("Problems loading image");
    vdev->sizeRead = sizeRead;
    vdev->imageReady = false;
  } else {
    PRINT_DEBUG (("VDev_getPPMImage - using pre-read image.\n"));
  }

  if (vdev->vid_buffer == NULL) return NULL;

  if (!(pixarray = (u_char*) malloc(vdev->image_w * vdev->image_h * 3 
				   * sizeof(u_char)))) {
    perror("Out of memory");
    return NULL;
  }
  
  idx = 0;
  for (row = 0; row < vdev->image_h; ++row) {
    for ( col = 0; col < vdev->image_w; ++col) {
      
      if (vdev->tmp_buffer != NULL) {
	val1 = vdev->tmp_buffer[idx+0];
	val2 = vdev->tmp_buffer[idx+1];
	val3 = vdev->tmp_buffer[idx+2];
      } else {
	val1 = vdev->vid_buffer[idx+0];
	val2 = vdev->vid_buffer[idx+1];
	val3 = vdev->vid_buffer[idx+2];
      }

      switch (format) {
      case RGB:
	pixarray[idx+0] = val1;
	pixarray[idx+1] = val2;
	pixarray[idx+2] = val3;
	break;
      case RBG:
	pixarray[idx+0] = val1;
	pixarray[idx+1] = val3;
	pixarray[idx+2] = val2;
	break;
      case GRB:
	pixarray[idx+0] = val2;
	pixarray[idx+1] = val1;
	pixarray[idx+2] = val3;
	break;
      case GBR:
	pixarray[idx+0] = val3;
	pixarray[idx+1] = val1;
	pixarray[idx+2] = val2;
	break;
      case BRG:
	pixarray[idx+0] = val2;
	pixarray[idx+1] = val3;
	pixarray[idx+2] = val1;
	break;
      case BGR:
	pixarray[idx+0] = val3;
	pixarray[idx+1] = val2;
	pixarray[idx+2] = val1;
	break;
      }      
      idx += 3;
    }
  }
  vdev->ppm_buffer = pixarray; 
  return vdev->ppm_buffer;

} /* end VDev_getPPMImage */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_getDescriptor: Return the file descriptor for the device
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: Device's file descriptor on success, or -1 if an error occurred.
 *
 */

extern int
VDev_getDescriptor(VDev_T vdev)
{
  PRINT_DEBUG (("VDev_getDescriptor\n"));

  if (vdev != NULL)
    if (vdev->camFd >= 0)
      return vdev->camFd;
  return VDEV_FAILURE;
} /* end VDev_getDescriptor */

/*
extern void eVDev_getPictInfo(VDev_T vdev, struct video_picture* vpic)
{
  if (vdev != NULL)
    vpic = &(vdev->vpic);
}
*/




/*----------------------------------------------------------------------------*/

/* 
 * VDev_getIntfName: Returns the name of the device through the V4L i/f, 
 *   like OV511 USB Camera
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: Device V4L name on success, or NULL if an error occurred.
 *
 */

extern char* 
VDev_getIntfName(VDev_T vdev)
{
  PRINT_DEBUG (("VDev_getIntfName\n"));

  if (vdev != NULL)
    return vdev->intfName;
  return NULL;
} /* end VDev_getIntfName */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_getNodeName: Returns the name of the device inode, eg /dev/video0
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: Device inode name on success, or NULL if an error occurred.
 *
 */

extern char* 
VDev_getNodeName(VDev_T vdev)
{
  PRINT_DEBUG (("VDev_getNodeName\n"));

  if (vdev != NULL)
    return vdev->nodeName;
  return NULL;
} /* end VDev_getNodeName */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_getRawSize: Returns size of last read
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: Size in bytes of the last read for image, or -1 if an error 
 *   occurred.
 *
 */

extern int
VDev_getRawSize(VDev_T vdev)
{
  PRINT_DEBUG (("VDev_getRawSize\n"));

  if (vdev != NULL)
    return vdev->sizeRead;
  return VDEV_FAILURE;
} /* end VDev_getRawSize */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_getMinSize: Returns minimum possible size
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: Zero and current minimum size on success, or -1 if an error 
 *   occurred.
 *
 */

extern int
VDev_getMinSize(VDev_T vdev, struct VDev_imageSize *size)
{
  PRINT_DEBUG (("VDev_getMinSize\n"));

  if (vdev != NULL || size != NULL) {
    size->width = vdev->min_w;
    size->height = vdev->min_h;
    return VDEV_SUCCESS;
  }
  return VDEV_FAILURE;
} /* end VDev_getMinSize */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_getMaxSize: Return maximum possible size
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: Zero and current max size on success, or -1 if an error 
 *   occurred.
 *
 */

extern int
VDev_getMaxSize(VDev_T vdev, struct VDev_imageSize *size)
{
  PRINT_DEBUG (("VDev_getMaxSize\n"));

  if (vdev != NULL || size != NULL) {
    size->width = vdev->max_w;
    size->height = vdev->max_h;
    return VDEV_SUCCESS;
  }
  return VDEV_FAILURE;
} /* end VDev_getMaxSize */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_getSize: Return image size (width, heigth)
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: Zero and current size on success, or -1 if an error occurred.
 *
 */

extern int
VDev_getSize(VDev_T vdev, struct VDev_imageSize *size)
{
  PRINT_DEBUG (("VDev_getSize\n"));

  if (vdev != NULL || size != NULL) {
    size->width = vdev->image_w;
    size->height = vdev->image_h;
    return VDEV_SUCCESS;
  }
  return VDEV_FAILURE;
} /* end VDev_getSize */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_setSize: This function will attempt to set a new image size; not all 
 *   sizes between the minimum and maximum size may be allowed by the device; 
 *   however, there is currently no way to retrieve a list of possible sizes. 
 *   It is safest to stick to CIF (352x288) and SIF (320x240) formats and 
 *   subsizes hereof, and VGA. Also it's wise to use to multiples of 8 in 
 *   both directions. SetSize will return zero when the size was rejected 
 *   by the driver.
 *
 * Parameters: vdev - handler for video device
 *             size - width and height
 * 
 * Returns: Zero on success, or -1 if an error occurred.
 *
 */

extern int
VDev_setSize(VDev_T vdev, struct VDev_imageSize size)
{
   struct video_window vwin;

   PRINT_DEBUG (("VDev_setSize (%d, %d)\n", size.width, size.height));
   
   if (vdev == NULL || vdev->camFd < 0 || 
       size.width > vdev->max_w || size.height > vdev->max_h)
     return VDEV_FAILURE;
   if (ioctl(vdev->camFd, VIDIOCGWIN, &vwin) < 0)
     return VDEV_FAILURE;

   MSync(vdev);

   vwin.width = size.width;
   vwin.height = size.height;
   vwin.clipcount = 0;
   if (ioctl(vdev->camFd, VIDIOCSWIN, &vwin) < 0) 
     return VDEV_FAILURE;
   
   /* 
    * Read back; it might be the hardware decides a bit differently 
    *  (for example, multiples of 8)
    *
    */
   if (ioctl(vdev->camFd, VIDIOCGWIN, &vwin) == 0) {
     vdev->image_w = vwin.width;
     vdev->image_h = vwin.height;
     if (vdev->image_w == 0 && vdev->image_h == 0) /* woops */
       return VDEV_FAILURE;
   }

   MCapture(vdev, vdev->curBuffer);
   
   if (vdev->vmbuf.size == 0) {
     free(vdev->vid_buffer); 
     vdev->image_buffer_size = vdev->image_w * vdev->image_h * 3; /* 4 | 3 ? */
     vdev->vid_buffer_size = vdev->image_buffer_size;
     vdev->vid_buffer = (u_char*) malloc(vdev->vid_buffer_size);
     if (!vdev->vid_buffer) {
       VDev_free(vdev);
       return VDEV_FAILURE;
     }
   }
   
   return VDEV_SUCCESS;
} /* end VDev_setSize */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_getFramerate: Return the current framerate.   
 *   This applies to some webcams that allow setting of a framerate.
 *   By default the framerate is set to 25 (assuming a videograbber card at
 *   PAL).
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: Framerate in frames per second on success, or -1 if an error 
 *   occurred.
 *
 */

extern int 
VDev_getFramerate(VDev_T vdev)
{
  struct video_window vwin;

  PRINT_DEBUG (("VDev_getFramerate\n"));

  if (vdev->camFd < 0)
    return -1;
  if (vdev->hasFramerate) {
    if (ioctl(vdev->camFd, VIDIOCGWIN, &vwin) < 0)
       return VDEV_FAILURE;
#if defined(PWC_FPS_SHIFT)       
     return (vwin.flags & PWC_FPS_FRMASK) >> PWC_FPS_SHIFT;
#endif     
   }
   return vdev->framerate;
} /* end VDev_getFramerate */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_setFramerate: Try to set a framerate.
 *   Some webcams allow their framerate to be set; this functions
 *   tries to do just that; in general, the camera will use the
 *   framerate closest to what it supports.
 *   Some cams can have a snapshot mode; the snap parameter enables
 *   this mode.
 *   If the cam does not support framerate setting, we have to do it
 *   ourselves and so we allow selection only from {1, 5, 10, 20, 40}
 *
 * Parameters: vdev - handler for video device
 *             fps  - framerate, in frames per second.
 *             snap - Use 'snapshot' mode
 *
 * Returns: Zero on success, or -1 if an error occurred.
 *
 */

extern int
VDev_setFramerate(VDev_T vdev, int fps, int snap)
{
  struct video_window vwin;

  PRINT_DEBUG(("VDev_setFramerate %d %d\n", fps, snap));

# if DEBUG
  printf(" [VDev_setFramerate(): fps=%d, snap=%d\n",fps,snap);
# endif

  if (vdev == NULL || vdev->camFd < 0)
    return VDEV_FAILURE;
  if (ioctl(vdev->camFd, VIDIOCGWIN, &vwin) < 0)
    return VDEV_FAILURE;
  if (vdev->hasFramerate) {

#   if DEBUG
    printf(" [VDev_setFramerate(): vdev has Framerate\n");
#   endif

    MSync(vdev);
    vdev->framerate = fps;
    if (vdev->framerate <= 0)
      vdev->framerate = 0;
    if (vdev->framerate > 63)
      vdev->framerate = 63;
#if defined(PWC_FPS_SHIFT)

    printf(" [VDev_setFramerate(): PWC_FPS_SHIFT defined\n");

    vwin.flags = (vwin.flags & ~PWC_FPS_MASK) | 
      (vdev->framerate << PWC_FPS_SHIFT);
    if (snap) vwin.flags |= PWC_FPS_SNAPSHOT;
#endif     
    if (ioctl(vdev->camFd, VIDIOCSWIN, &vwin) < 0)
      return VDEV_FAILURE;
    MCapture(vdev, vdev->curBuffer);
  } else {
    /* ugly but effective - fps restricted to {1, 5, 10, 20, 40} */
    vdev->framerate = fps;
#   if DEBUG
    printf(" [VDev_setFramerate(): vdev->framerate=%d\n",vdev->framerate);
#   endif
    if (vdev->framerate <= 0) vdev->framerate = 1;
    /*vvv*/
    /*
    if (vdev->framerate > 1  || vdev->framerate < 10) vdev->framerate = 5;
    if (vdev->framerate > 5  || vdev->framerate < 11) vdev->framerate = 10;
    if (vdev->framerate > 10 || vdev->framerate < 21) vdev->framerate = 20;
    */
    if (vdev->framerate > 1  && vdev->framerate < 10) vdev->framerate = 5;
    if (vdev->framerate > 5  && vdev->framerate < 11) vdev->framerate = 10;
    if (vdev->framerate > 10 && vdev->framerate < 21) vdev->framerate = 20;
    /*^^^*/

    if (vdev->framerate > 20)
      vdev->framerate = 40;
  }

# if DEBUG
  printf(" [VDev_setFramerate(): vdev->framerate=%d, snap=%d\n",vdev->framerate,snap);
# endif

  return VDEV_SUCCESS;
} /* end VDev_setFramerate */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_getBrightness: Return brightness setting of device.
 *    
 * Parameters: vdev - handler for video device
 * 
 * Returns: Brightness, an nsigned integer in [0,65535], or -1 if an error
 *   occurred. 65535 may indicate setting is not available.
 *
 */

extern u_int
VDev_getBrightness(VDev_T vdev)
{
  PRINT_DEBUG (("VDev_getBrightness\n"));

  if (vdev != NULL)
    return vdev->vpic.brightness;
  return VDEV_FAILURE;
} /* end VDev_getBrightness */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_setBrightness: Set brightness in device. The value returned by 
 *   getBrightness may be slightly different from what is set with 
 *   setBrightness.
 *
 * Parameters: vdev - handler for video device
 *             val - Brightness, an integer in [0, 65535].
 *
 * Returns: Zero on success, or -1 if an error occurred.
 *
 */

extern int
VDev_setBrightness(VDev_T vdev, u_int val)
{
  PRINT_DEBUG (("VDev_setBrightness %u\n", val));

  if (vdev == NULL || vdev->camFd < 0)
    return VDEV_FAILURE;
  vdev->vpic.brightness = val & 0xffff;
  if (ioctl(vdev->camFd, VIDIOCSPICT, &(vdev->vpic)) < 0)
    return VDEV_FAILURE;
  return VDEV_SUCCESS;
} /* end VDev_setBrightness */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_getContrast: Return contrast setting of device.
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: Constrast, an unsigned integer in [0,65535], or -1 if an error
 *   occurred. 65535 may indicate setting is not available.
 *
 */

extern u_int
VDev_getContrast(VDev_T vdev)
{
  PRINT_DEBUG (("VDev_getContrast\n"));

  if (vdev != NULL)
    return vdev->vpic.contrast;
  return VDEV_FAILURE;
} /* end VDev_getContrast */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_setContrast: Set contrast in device.
 *   The value returned by getContrast may be slightly different from what is
 *   set with setContrast.
 *
 * Parameters: vdev - handler for video device
 *             val  - Contrast, an integer in [0,65535].
 * 
 * Returns: Zero on success, or -1 if an error occurred.
 *
 */

extern int
VDev_setContrast(VDev_T vdev, u_int val)
{
  PRINT_DEBUG (("VDev_setContrast %u\n", val));

  if (vdev == NULL || vdev->camFd < 0)
    return VDEV_FAILURE;
  vdev->vpic.contrast = val & 0xffff;
  if (ioctl(vdev->camFd, VIDIOCSPICT, &(vdev->vpic)) < 0)
    return VDEV_FAILURE;
  return VDEV_SUCCESS;
} /* end VDev_setContrast */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_getHue: Return hue (color shift) setting of device.
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: Hue, an unsigned integer in [0,65535], or -1 if an error 
 *   occurred. 65535 may indicate setting is not available.
 *
 */

extern u_int
VDev_getHue(VDev_T vdev)
{
  PRINT_DEBUG (("VDev_getHue\n"));

  if (vdev != NULL)
    return vdev->vpic.hue;
  return VDEV_FAILURE;
} /* end VDev_getHue */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_setHue: Set hue in device. Hue is a way to correct for colour
 *   deviations. The value returned by getHue may be slightly different from
 *   what is set with setHue.
 *
 * Parameters: vdev - handler for video device
 *             val - Hue, an unsigned integer in [0,65535].
 *
 * Returns: Zero on success, or -1 if an error occurred.
 *
 */

extern int
VDev_setHue(VDev_T vdev, u_int val)
{
  PRINT_DEBUG (("VDev_setHue %u\n", val));

  if (vdev == NULL || vdev->camFd < 0)
      return VDEV_FAILURE;
  vdev->vpic.hue = val & 0xffff;
  if (ioctl(vdev->camFd, VIDIOCSPICT, &(vdev->vpic)) < 0)
    return VDEV_FAILURE;
  return VDEV_SUCCESS;
} /* end VDev_setHue */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_getColor: Return colour saturation setting of device.
 *   A colour saturation of 0 means no colour at all, so the returned images
 *   are grayscale.
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: Color, an unsigned integer in [0,65535], or -1 if an error 
 *   occurred. 65535 may indicate setting is not available. 
 *
 */

extern u_int
VDev_getColor(VDev_T vdev)
{
  PRINT_DEBUG (("VDev_getColor\n"));

  if (vdev != NULL)
    return vdev->vpic.colour;
  return VDEV_FAILURE;
} /* end VDev_getColor */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_setColor: Set colour saturation in device. Colour saturation sets 
 *   how bright colours should appear. A saturation of 0 yields grayscale 
 *   images.
 *   The value returned by getColour may be slightly different from what 
 *   is set with setColour.
 * 
 * Parameters: vdev - handler for video device
 *             val  - Color, an unsigned integer in [0,65535].
 *
 * Returns: Zero on success, or -1 if an error occurred.
 *
 */

extern int
VDev_setColor(VDev_T vdev, u_int val)
{
  PRINT_DEBUG (("VDev_setColor %u\n", val));

  if (vdev == NULL || vdev->camFd < 0)
      return VDEV_FAILURE;
   vdev->vpic.colour = val & 0xffff;
   if (ioctl(vdev->camFd, VIDIOCSPICT, &(vdev->vpic)) < 0)
     return VDEV_FAILURE;
   return VDEV_SUCCESS;
} /* end VDev_setColor */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_getWhiteness: Return gamma setting of device. Sometimes used as a 
 *   brightness contrast, but more generally this returns the gamma 
 *   correction the device applies to the image.
 *
 * Parameters: vdev - handler for video device
 * 
 * Returns: Whiteness, an unsigned integer in [0,65535], or -1 if an error 
 *   occurred. 65535 may indicate setting is not available.
 *
 */

extern u_int
VDev_getWhiteness(VDev_T vdev)
{
  PRINT_DEBUG (("VDev_getWhiteness\n"));

  if (vdev != NULL)
    return vdev->vpic.whiteness;
  return VDEV_FAILURE;
} /* end VDev_getWhiteness */


/*----------------------------------------------------------------------------*/

/* 
 * VDev_setWhiteness: Set gamma value in device. The value returned by 
 *   getWhiteness may be slightly different from what is set with 
 *   SetWhiteness.
 *   Whiteness is sometimes used as brightness, but usually this sets the
 *   gamma correction the device will apply to the image.
 *
 * Parameters: vdev - handler for video device
 *             val  - whiteness, an integer in [0,65535].
 * 
 * Returns: Zero on success, or -1 if an error occurred.
 *
 */

extern int
VDev_setWhiteness(VDev_T vdev, u_int val)
{
  PRINT_DEBUG (("VDev_setWhiteness %u\n", val));

  if (vdev == NULL || vdev->camFd < 0)
      return VDEV_FAILURE;
  vdev->vpic.whiteness = val & 0xffff;
  if (ioctl(vdev->camFd, VIDIOCSPICT, &(vdev->vpic)) < 0)
     return VDEV_FAILURE;
   return VDEV_SUCCESS;
} /* end VDev_setWhiteness */


/*----------------------------------------------------------------------------*/

