/* -*-C-*-
*******************************************************************************
*
* File:         vdevTest.c
* RCS:          $Id: vdevTest.c,v 1.2 2001/11/05 20:03:57 fabianb Exp $
* Description:  Test for libvdev.a: a library to access USB video cams 
*               in Linux.
* Author:       Fabian E. Bustamante
*               Systems Research Group
*               College of Computing
*               Georgia Institute of Technology
* Created:      Sun May 13, 2001 at 10:28:14
* Modified:     Mon Nov 05, 2001 at 14:55:19 fabianb@cc.gatech.edu
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
* (C) Copyright 2001, Georgia Institute of Technology, all rights reserved.
*
*******************************************************************************
*/

#if HAVE_CONFIG_H
#  include <src/config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>      
#include <getopt.h>
#include <vdev.h>

typedef enum {FALSE, TRUE} boolean;

static char* local_progname;   
static boolean verbose = FALSE;

/* Command line arguments */
static struct option longopts[] =
{
  {"verbose", no_argument, NULL, 'v'},
  {"help", no_argument, NULL, 'h'},
  {"nframes", required_argument, NULL, 'n'},
  {"format", required_argument, NULL, 'f'},
  {NULL, no_argument, NULL, 0}
};


/* 
 * PPM stuff
 *
 */

typedef unsigned char pixval;

typedef struct {
  pixval r, g, b;
} pixel;


#define PPM_OVERALLMAXVAL 65535
#define PPM_MAXMAXVAL 255

#define PPM_MAGIC1 'P'
#define PPM_MAGIC2 '3'
#define RPPM_MAGIC2 '6'
#define PPM_FORMAT (PPM_MAGIC1 * 256 + PPM_MAGIC2)
#define RPPM_FORMAT (PPM_MAGIC1 * 256 + RPPM_MAGIC2)
#define PPM_TYPE PPM_FORMAT

#define PPM_ASSIGN(p,red,grn,blu) \
   do { (p).r = (red); (p).g = (grn); (p).b = (blu); } while ( 0 )

/*----------------------------------------------------------------------------*/

/*
 * usage: Prints out 'usage' string
 *
 * Parameters: target 'stream' and 'status' to quit with
 *
 * Returns: -
 *
 */
 
static void
usage (FILE* stream, int status)
{
  fprintf(stream, "Usage: %s [-vh] -n num-frames [-f format]\n"
	          "\t -v --verbose            \t set verbose mode\n"
	          "\t -h --help               \t give this help\n"
	          "\t -n --nframes num-frames \t number of images\n"
	          "\t -f --format  format     \t RGB format (dflt 2, RGB)\n", 
	  local_progname);
  exit(status);
} /* end usage */

/*----------------------------------------------------------------------------*/

int
main (int argc, char* argv[])
{
  VDev_T vdev;
  char snapshotFile[40];
  int shotfd; 
  char header[20];
  char pixelline[3];
  int rawsize;
  struct tm *now;
  time_t nowt;
  pixel *pixrow, *pP;
  u_char* image;
  u_char val1, val2, val3;
  int i, j, idx, row, col;
  int numFrames = 0;
  int format = RBG;
  struct VDev_imageSize size;
  int optc;

  local_progname = argv[0];
  while ((optc = getopt_long(argc, argv, "+vhn:f:", 
			     longopts, (int*) 0)) != EOF) {
    switch (optc) {
    case 'v':
      verbose = TRUE;
      break;
    case 'h':
      usage(stdout, 0);
      break;
    case 'n':
      numFrames = atoi(optarg);
      if (numFrames < 0) usage (stderr, 1);
      break; 
    case 'f':
      format = atoi(optarg);
      if (format < 1) usage(stderr, 1);
      break;
    default:
      usage(stderr, 1);
    }
  }
  
  if (numFrames < 1) usage(stderr, 1);

  /* no extra args */
  if (optind > argc) usage(stderr, 1);

  VDev_init(320,240);
  if (verbose) printf("Num devices: %d\n", VDev_numVDevices());

  vdev = VDev_getVDevice(0);
  if (vdev == NULL) fprintf(stderr, "Wrong device?\n");
  if (verbose) {
    printf("Valid? %c\n", VDev_isValid(vdev) ? 'T':'F');
    printf("V4L name:    %s\n", VDev_getIntfName(vdev));
    printf("Device name: %s\n", VDev_getNodeName(vdev));
    VDev_getMaxSize(vdev, &size);
    printf("Max size (width, height): (%d, %d)\n", size.width, size.height);
    VDev_getMinSize(vdev, &size);
    printf("Min size (width, height): (%d, %d)\n", size.width, size.height);   
  }

  VDev_getSize(vdev, &size);
  if (verbose) printf("Size (width, heigth): (%d, %d)\n", size.width, size.height);
  
  VDev_open(vdev, 1);
  if (verbose) {
    printf("File descriptor: %d\n", VDev_getDescriptor(vdev));  
    printf("Frame rate: %d\n", VDev_getFramerate(vdev));
    printf("Brightness: %u\n", VDev_getBrightness(vdev));
    printf("Contrast: %u\n", VDev_getContrast(vdev));
    printf("Hue: %u\n", VDev_getHue(vdev));
    printf("Color: %u\n", VDev_getColor(vdev));
    printf("Whiteness: %u\n", VDev_getWhiteness(vdev));
  }
  
  for (i = 0; i < numFrames; i++) {
    printf("Ready for another picture? [ENTER]\n");
    getchar();  

    time(&nowt);
    now = localtime(&nowt);

    /* Raw image file */
    sprintf(snapshotFile, "snapshot-raw%d-%d-%d-%d", 
	    i+1, now->tm_hour, now->tm_min, now->tm_sec);
    if ((shotfd = open(snapshotFile, O_CREAT|O_WRONLY, 
		       S_IRWXU|S_IRGRP|S_IROTH)) < 0) {
      perror("Problems opening raw snapshot file");
      exit(1);
    }
    image = VDev_getImage(vdev);
    rawsize = VDev_getRawSize(vdev);
    printf ("RAWSIZE %d\n", rawsize);

    write(shotfd, image, rawsize);
    close(shotfd);

    /* PPM image file locally derived from our raw image */
    sprintf(snapshotFile, "snapshot-derived%d-%d-%d-%d.ppm", 
	    i+1, now->tm_hour, now->tm_min, now->tm_sec);

    if ((shotfd = open(snapshotFile, O_CREAT|O_WRONLY, 
		       S_IRWXU|S_IRGRP|S_IROTH)) < 0) {
      perror("Problems opening PPM-derived snapshot file");
      exit(1);
    }

    sprintf(header, "%c%c\n%d %d\n%d\n", PPM_MAGIC1, RPPM_MAGIC2,  
	    size.width, size.height, 255);  
    write(shotfd, header, sizeof(header));
    
    if (!(pixrow = (pixel*)malloc(size.width * sizeof(pixel)))) {
      perror("Out of memory");
      exit(1);
    }
    
    idx = 0;
    for ( row = 0; row < size.height; ++row) {
      for ( col = 0, pP = pixrow; col < size.width; ++col, ++pP ) {
	val1 = image[idx+0];
	val2 = image[idx+1];
	val3 = image[idx+2];
	
	/* PPM_ASSIGN( *pP, val1, val2, val3 ); RGB */
	PPM_ASSIGN( *pP, val1, val3, val2 ); /* RBG */
	/* PPM_ASSIGN( *pP, val2, val1, val3 ); GRB */
	/* PPM_ASSIGN( *pP, val2, val3, val1 ); BRG */
	/* PPM_ASSIGN( *pP, val3, val1, val2 ); GBR */
	/* PPM_ASSIGN( *pP, val3, val2, val1 ); BGR */
	
	idx += 3;
      }

      for (j = 0, pP = pixrow; j < size.width; ++j, ++pP) {
	sprintf(pixelline, "%c%c%c", pP->r, pP->g, pP->b);
	write(shotfd, pixelline, 3);
      }
    }

    close(shotfd);
    
    /* PPM image file as given by VDev */
    sprintf(snapshotFile, "snapshot%d-%d-%d-%d.ppm", 
	    i+1, now->tm_hour, now->tm_min, now->tm_sec);
    
    if ((shotfd = open(snapshotFile, O_CREAT|O_WRONLY, 
		       S_IRWXU|S_IRGRP|S_IROTH)) < 0) {
      perror("Problems opening PPM snapshot file");
      exit(1);
    }
    
    sprintf(header, "%c%c\n%d %d\n%d\n", PPM_MAGIC1, RPPM_MAGIC2,  
	    size.width, size.height, 255);  
    write(shotfd, header, sizeof(header));
    image = VDev_getPPMImage(vdev, format);
    write(shotfd, image, rawsize);
    close(shotfd);
  }
  
  
  VDev_close(vdev);

  VDev_finalize();
  exit(0);
} /* end main */
